"""Functions for reading and writing data."""

import io
import json
import logging
import os
import pickle as pkl
import re
import struct
from copy import deepcopy
from datetime import datetime
from pathlib import Path
from typing import Any

import h5py
import numpy as np
import numpy.typing as npt
import pandas as pd
from AFMReader import asd, gwy, ibw, jpk, spm, stp, top, topostats  # pylint: disable=no-name-in-module
from numpyencoder import NumpyEncoder
from packaging.version import parse as parse_version
from ruamel.yaml import YAML, YAMLError

from topostats import CONFIG_DOCUMENTATION_REFERENCE, TOPOSTATS_BASE_VERSION, TOPOSTATS_COMMIT, __release__, __version__
from topostats.classes import (
    DisorderedTrace,
    GrainCrop,
    MatchedBranch,
    Molecule,
    Node,
    OrderedTrace,
    TopoStats,
    convert_to_dict,
)
from topostats.logs.logs import LOGGER_NAME

LOGGER = logging.getLogger(LOGGER_NAME)

# pylint: disable=broad-except
# pylint: disable=too-many-lines
# pylint: disable=too-many-branches


# Sylvia: Ruff says too complex but I think breaking this out would be more complex.
# pylint: disable=too-many-return-statements
def dict_almost_equal(dict1: dict, dict2: dict, abs_tol: float = 1e-9):  # noqa: C901
    """
    Recursively check if two dictionaries are almost equal with a given absolute tolerance.

    This should really just be iterative and is an affront to memory usage.

    Parameters
    ----------
    dict1 : dict
        First dictionary to compare.
    dict2 : dict
        Second dictionary to compare.
    abs_tol : float
        Absolute tolerance to check for equality.

    Returns
    -------
    bool
        True if the dictionaries are almost equal, False otherwise.
    """
    # Ensure both dictionaries share the same keys
    if dict1.keys() != dict2.keys():
        return False

    # Ensure the types of the values are the same
    for key in dict1:
        # Sylvia: Pylint complains about calling the type() function in this way, but it is the only way to
        # ensure that the types of the values in both dictionaries are the same (that I know of).
        # Replace with better way if you know of one.
        # pylint: disable=unidiomatic-typecheck
        if type(dict1[key]) != type(dict2[key]):  # noqa: E721
            LOGGER.debug(f"Key {key} types not equal: {type(dict1[key])} != {type(dict2[key])}")
            return False

    LOGGER.debug("Comparing dictionaries")

    for key in dict1:
        LOGGER.debug(f"Comparing key {key}")
        if isinstance(dict1[key], dict):
            if not dict_almost_equal(dict1[key], dict2[key], abs_tol=abs_tol):
                return False
        elif isinstance(dict1[key], np.ndarray):
            if not np.allclose(dict1[key], dict2[key], atol=abs_tol):
                LOGGER.debug(f"Key {key} type: {type(dict1[key])} not equal: {dict1[key]} != {dict2[key]}")
                return False
        elif isinstance(dict1[key], float):
            # Skip if both values are NaN
            if not (np.isnan(dict1[key]) and np.isnan(dict2[key])):
                # Check if both values are close
                if not np.isclose(dict1[key], dict2[key], atol=abs_tol):
                    LOGGER.debug(f"Key {key} type: {type(dict1[key])} not equal: {dict1[key]} != {dict2[key]}")
                    return False
        elif isinstance(dict1[key], list):
            if not lists_almost_equal(dict1[key], dict2[key], abs_tol=abs_tol):
                return False
        elif dict1[key] != dict2[key]:
            LOGGER.debug(f"Key {key} not equal: {dict1[key]} != {dict2[key]}")
            return False

    return True


def lists_almost_equal(list1: list, list2: list, abs_tol: float = 1e-9) -> bool:
    """
    Check if two lists are almost equal with a given absolute tolerance.

    Note: Currently the lists must be flat, the same length and contain only numbers (int or float).

    Parameters
    ----------
    list1 : list
        First list to compare.
    list2 : list
        Second list to compare.
    abs_tol : float
        Absolute tolerance to check for equality.

    Returns
    -------
    bool
        True if the lists are almost equal, False otherwise.

    Raises
    ------
    NotImplementedError
        If the items in the lists are not of type int or float.
    """
    # Check if both lists are the same length and only contain numbers
    if len(list1) != len(list2):
        LOGGER.debug(f"Lists not same length: {len(list1)} != {len(list2)}")
        return False
    for i, (item1, item2) in enumerate(zip(list1, list2)):
        if isinstance(item1, int | float | np.int64) and isinstance(item2, int | float | np.int64):
            if not np.isclose(item1, item2, atol=abs_tol):
                LOGGER.debug(f"List item {i} not equal: {item1} != {item2}")
                return False
        elif isinstance(item1, list) and isinstance(item2, list):
            if not lists_almost_equal(item1, item2):
                LOGGER.debug(f"Nested list item {i} not equal: {item1} != {item2}")
                return False
        else:
            raise NotImplementedError(
                f"Comparison of list items of type {type(item1)} inside a list is not implemented."
            )
    return True


def read_yaml(filename: str | Path) -> dict:
    """
    Read a YAML file.

    Parameters
    ----------
    filename : Union[str, Path]
        YAML file to read.

    Returns
    -------
    Dict
        Dictionary of the file.
    """
    with Path(filename).open(encoding="utf-8") as f:
        try:
            yaml_file = YAML(typ="safe")
            return yaml_file.load(f)
        except YAMLError as exception:
            LOGGER.error(exception)
            return {}


def get_date_time() -> str:
    """
    Get a date and time for adding to generated files or logging.

    Returns
    -------
    str
        A string of the current date and time, formatted appropriately.
    """
    return datetime.now().strftime("%Y-%m-%d %H:%M:%S")


def write_yaml(
    config: dict,
    output_dir: str | Path,
    config_file: str = "config.yaml",
    header_message: str = None,
) -> None:
    """
    Write a configuration (stored as a dictionary) to a YAML file.

    Parameters
    ----------
    config : dict
        Configuration dictionary.
    output_dir : Union[str, Path]
        Path to save the dictionary to as a YAML file (it will be called 'config.yaml').
    config_file : str
        Filename to write to.
    header_message : str
        String to write to the header message of the YAML file.
    """
    # Save the configuration to output directory
    output_config = Path(output_dir) / config_file
    # Revert PosixPath items to string
    config = path_to_str(config)

    if header_message:
        header = f"# {header_message} : {get_date_time()}\n" + CONFIG_DOCUMENTATION_REFERENCE
    else:
        header = f"# Configuration from TopoStats run completed : {get_date_time()}\n" + CONFIG_DOCUMENTATION_REFERENCE

    # Add comment to config with topostats version + commit
    header += f"# TopoStats version: {TOPOSTATS_BASE_VERSION}\n"
    header += f"# Commit: {TOPOSTATS_COMMIT}\n"

    output_config.write_text(header, encoding="utf-8")

    yaml = YAML(typ="safe")
    with output_config.open("a", encoding="utf-8") as f:
        try:
            yaml.dump(config, f)
        except YAMLError as exception:
            LOGGER.error(exception)


def save_array(array: npt.NDArray, outpath: Path, filename: str, array_type: str) -> None:
    """
    Save a Numpy array to disk.

    Parameters
    ----------
    array : npt.NDArray
        Numpy array to be saved.
    outpath : Path
        Location array should be saved.
    filename : str
        Filename of the current image from which the array is derived.
    array_type : str
        Short string describing the array type e.g. z_threshold. Ideally should not have periods or spaces in (use
        underscores '_' instead).
    """
    np.save(outpath / f"{filename}_{array_type}.npy", array)
    LOGGER.info(f"[{filename}] Numpy array saved to : {outpath}/{filename}_{array_type}.npy")


def load_array(array_path: str | Path) -> npt.NDArray:
    """
    Load a Numpy array from file.

    Should have been saved using save_array() or numpy.save().

    Parameters
    ----------
    array_path : Union[str, Path]
        Path to the Numpy array on disk.

    Returns
    -------
    npt.NDArray
        Returns the loaded Numpy array.
    """
    try:
        return np.load(Path(array_path))
    except FileNotFoundError as e:
        raise e


def path_to_str(config: dict) -> dict:
    """
    Recursively traverse a dictionary and convert any Path() objects to strings for writing to YAML.

    Parameters
    ----------
    config : dict
        Dictionary to be converted.

    Returns
    -------
    Dict:
        The same dictionary with any Path() objects converted to string.
    """
    for key, value in config.items():
        if isinstance(value, dict):
            path_to_str(value)
        elif isinstance(value, Path):
            config[key] = str(value)

    return config


def get_out_path(image_path: str | Path = None, base_dir: str | Path = None, output_dir: str | Path = None) -> Path:
    """
    Add the image path relative to the base directory to the output directory.

    Parameters
    ----------
    image_path : Path
        The path of the current image.
    base_dir : Path
        Directory to recursively search for files.
    output_dir : Path
        The output directory specified in the configuration file.

    Returns
    -------
    Path
        The output path that mirrors the input path structure.
    """
    # If image_path is relative and doesn't include base_dir then a ValueError is raised, in which
    # case we just want to append the image_path to the output_dir
    image_path = Path(image_path)
    try:
        # Remove the filename if there is a suffix, not always the case as
        # get_out_path is called from save_folder_grainstats()
        if image_path.suffix:
            return output_dir / image_path.relative_to(base_dir).parent / image_path.parts[-1]
        return output_dir / image_path.relative_to(base_dir)
    except ValueError:
        if image_path.suffix:
            return output_dir / image_path.parent / image_path.parts[-1]
        return Path(str(output_dir) + "/" + str(image_path))


def find_files(base_dir: str | Path = None, file_ext: str = ".spm") -> list:
    """
    Recursively scan the specified directory for images with the given file extension.

    Parameters
    ----------
    base_dir : Union[str, Path]
        Directory to recursively search for files, if not specified the current directory is scanned.
    file_ext : str
        File extension to search for.

    Returns
    -------
    List
        List of files found with the extension in the given directory.
    """
    base_dir = Path("./") if base_dir is None else Path(base_dir)
    if file_ext == ".spm":
        return list(base_dir.glob("**/*" + file_ext)) + _find_old_bruker_files(base_dir)
    return list(base_dir.glob("**/*" + file_ext))


def _find_old_bruker_files(base_dir: Path) -> list[Path]:
    r"""
    Find old Bruker files that have old extensions.

    Older Bruker files have extensions such as ``.001``, ``.002`` rather than ``.spm``. AFMReader can handle these fine
    but TopoStats needs to include them when finding files.

    Parameters
    ----------
    base_dir : Path
        Directory to search for files.

    Returns
    -------
    list[Path]
        List of files that match the regex ``\.\d+$``.
    """
    # Compile regex to match files ending with a period and one or more digits. List comprehension of matches is
    # returned.
    OLD_BRUKER_RE = re.compile(r"\.\d+$")
    return [p for p in base_dir.glob("**/*") if OLD_BRUKER_RE.match(p.suffix)]


def save_image_grainstats(
    output_dir: str | Path, base_dir: str | Path, all_stats_df: pd.DataFrame, stats_filename: str
) -> None:
    """
    Save a data frame of grain and tracing statistics at the folder level.

    Parameters
    ----------
    output_dir : Union[str, Path]
        Path of the output directory head.
    base_dir : Union[str, Path]
        Path of the base directory where files were found.
    all_stats_df : pd.DataFrame
        The dataframe containing all sample statistics run.
    stats_filename : str
        The name of the type of statistics dataframe to be saved.

    Returns
    -------
    None
        This only saves the dataframes and does not retain them.
    """
    dirs = set(all_stats_df["basename"].values)
    if output_dir.stem != "processed":
        output_dir_processed = Path(str(output_dir)) / "processed"
        output_dir_processed.mkdir(parents=True, exist_ok=True)
    LOGGER.debug(f"Statistics :\n{all_stats_df}")
    for _dir in dirs:
        LOGGER.debug(f"Statistics ({_dir}) :\n{all_stats_df}")
        try:
            # Ensure "output/processed" directory exists at the stem of out_path, creating if needed
            out_path = get_out_path(Path(_dir), base_dir, output_dir_processed)
            out_path.mkdir(parents=True, exist_ok=True)
            all_stats_df[all_stats_df["basename"] == _dir].to_csv(out_path / f"image_{stats_filename}.csv", index=True)
            LOGGER.info(f"Image-wise statistics saved to: {str(out_path)}/image_{stats_filename}.csv")
        except TypeError:
            LOGGER.info(f"No image-wise statistics for image {_dir}, no grains detected.")


def read_null_terminated_string(open_file: io.TextIOWrapper, encoding: str = "utf-8") -> str:
    """
    Read an open file from the current position in the open binary file, until the next null value.

    Parameters
    ----------
    open_file : io.TextIOWrapper
        An open file object.
    encoding : str
        Encoding to use when decoding the bytes.

    Returns
    -------
    str
        String of the ASCII decoded bytes before the next null byte.

    Examples
    --------
    >>> with open("test.txt", "rb") as f:
    ...     print(read_null_terminated_string(f), encoding="utf-8")
    """
    byte = open_file.read(1)
    value = b""
    while byte != b"\x00":
        value += byte
        byte = open_file.read(1)
    # Sometimes encodings cannot decode a byte that is not defined in the encoding.
    # Try 'latin1' in this case as it is able to handle symbols such as micro (µ).
    try:
        return str(value.decode(encoding=encoding))
    except UnicodeDecodeError as e:
        if "codec can't decode byte" in str(e):
            bad_byte = str(e).split("byte ")[1].split(":")[0]
            LOGGER.debug(
                f"Decoding error while reading null terminated string. Encoding {encoding} encountered"
                f" a byte that could not be decoded: {bad_byte}. Trying 'latin1' encoding."
            )
            return str(value.decode(encoding="latin1"))
        raise e


def read_u32i(open_file: io.TextIOWrapper) -> str:
    """
    Read an unsigned 32 bit integer from an open binary file (in little-endian form).

    Parameters
    ----------
    open_file : io.TextIOWrapper
        An open file object.

    Returns
    -------
    int
        Python integer type cast from the unsigned 32 bit integer.
    """
    return int(struct.unpack("<i", open_file.read(4))[0])


def read_64d(open_file: io.TextIOWrapper) -> str:
    """
    Read a 64-bit double from an open binary file.

    Parameters
    ----------
    open_file : io.TextIOWrapper
        An open file object.

    Returns
    -------
    float
        Python float type cast from the double.
    """
    return float(struct.unpack("d", open_file.read(8))[0])


def read_char(open_file: io.TextIOWrapper) -> str:
    """
    Read a character from an open binary file.

    Parameters
    ----------
    open_file : io.TextIOWrapper
        An open file object.

    Returns
    -------
    str
        A string type cast from the decoded character.
    """
    return open_file.read(1).decode("ascii")


def read_gwy_component_dtype(open_file: io.TextIOWrapper) -> str:
    """
    Read the data type of a `.gwy` file component.

    Possible data types are as follows:

    - 'b': boolean
    - 'c': character
    - 'i': 32-bit integer
    - 'q': 64-bit integer
    - 'd': double
    - 's': string
    - 'o': `.gwy` format object

    Capitalised versions of some of these data types represent arrays of values of that data type. Arrays are stored as
    an unsigned 32 bit integer, describing the size of the array, followed by the unseparated array values:

    - 'C': array of characters
    - 'I': array of 32-bit integers
    - 'Q': array of 64-bit integers
    - 'D': array of doubles
    - 'S': array of strings
    - 'O': array of objects.

    Parameters
    ----------
    open_file : io.TextIOWrapper
        An open file object.

    Returns
    -------
    str
        Python string (one character long) of the data type of the component's value.
    """
    return open_file.read(1).decode("ascii")


def get_relative_paths(paths: list[Path]) -> list[str]:
    """
    Extract a list of relative paths, removing the common suffix.

    From a list of paths, create a list where each path is relative to all path's closest common parent. For example,
    ['a/b/c', 'a/b/d', 'a/b/e/f'] would return ['c', 'd', 'e/f'].

    Parameters
    ----------
    paths : list
        List of string or pathlib paths.

    Returns
    -------
    list
        List of string paths, relative to the common parent.
    """
    # Ensure paths are all pathlib paths, and not strings
    paths = [Path(path) for path in paths]

    # If the paths list consists of all the same path, then the relative path will
    # be '.', which we don't want. we want the relative path to be the full path probably.
    # len(set(my_list)) == 1 determines if all the elements in a list are the same.
    if len(set(paths)) == 1:
        return [str(path.as_posix()) for path in paths]

    deepest_common_path = os.path.commonpath(paths)
    # Have to convert to strings else the dataframe values will be slightly different
    # to what is expected.
    return [str(path.relative_to(deepest_common_path).as_posix()) for path in paths]


def convert_basename_to_relative_paths(df: pd.DataFrame):
    """
    Convert paths in the 'basename' column of a dataframe to relative paths.

    If the 'basename' column has the following paths: ['/usr/topo/data/a/b', '/usr/topo/data/c/d'], the output will be:
    ['a/b', 'c/d'].

    Parameters
    ----------
    df : pd.DataFrame
        A pandas dataframe containing a column 'basename' which contains the paths
        indicating the locations of the image data files.

    Returns
    -------
    pd.DataFrame
        A pandas dataframe where the 'basename' column has paths relative to a common
        parent.
    """
    paths = df["basename"].tolist()
    paths = [Path(path) for path in paths]
    relative_paths = get_relative_paths(paths=paths)
    df["basename"] = relative_paths

    return df


# pylint: disable=too-many-instance-attributes
class LoadScans:
    """
    Load the image and image parameters from a file path.

    Parameters
    ----------
    img_paths : list[str, Path]
        Path to a valid AFM scan to load.
    config : dict[str, Any]
        Dictionary of all configuration options.
    channel : str
        Image channel to extract from the scan.
    """

    def __init__(
        self,
        img_paths: list[str | Path],
        config: dict[str, Any],
        channel: str | None = None,
    ):
        """
        Initialise the class.

        Parameters
        ----------
        img_paths : list[str | Path]
            Path to a valid AFM scan to load.
        config : dict[str, Any]
            Dictionary of all configuration options.
        channel : str
            Image channel to extract from the scan.
        """
        self.img_paths = img_paths
        self.img_path = None
        self.channel = config["loading"]["channel"] if channel is None else channel
        self.channel_data = None
        self.filename = None
        self.suffix = None
        self.image = None
        self.pixel_to_nm_scaling = None
        self.grain_masks = {}
        self.grain_trace_data = {}
        self.img_dict: dict[str, TopoStats] = {}
        self.MINIMUM_IMAGE_SIZE = 10
        self.config = config

    def load_spm(self) -> tuple[npt.NDArray, float]:
        """
        Extract image and pixel to nm scaling from the Bruker .spm file.

        Returns
        -------
        tuple[npt.NDArray, float]
            A tuple containing the image and its pixel to nanometre scaling value.
        """
        try:
            LOGGER.debug(f"Loading image from : {self.img_path}")
            return spm.load_spm(file_path=self.img_path, channel=self.channel)
        except FileNotFoundError:
            LOGGER.error(f"File Not Found : {self.img_path}")
            raise

    def load_topostats(self) -> dict[str, Any]:
        """
        Load a ``.topostats`` file (hdf5 format) using AFMReader.

        AFMReader is general and returns the data as a dictionary. This is converted to ``TopoStasts`` class later when
        building dictionaries of images.

        Returns
        -------
        dict[str, Any]
            A dictionary of all previously processed data and configuration options.
        """
        try:
            LOGGER.debug(f"Loading image from : {self.img_path}")
            raw_data = topostats.load_topostats(self.img_path)
            if "topostats_file_version" in raw_data.keys():  # pylint: disable=consider-iterating-dictionary
                LOGGER.warning(
                    f"[{raw_data['filename']}] : This '.topostats' is an old format "
                    f"({raw_data['topostats_file_version']}), only core features are loaded. "
                    "All trace data has been dropped. If you need access to these please use AFMReader directly."
                )
                # Remove trace data, masks and unused variables, may not have all attributes
                old_keys_to_remove = [
                    "disordered_traces",
                    "grain_curvature_stats",
                    "grain_masks",
                    "grain_masks",
                    "grain_trace_data",
                    "heigh_profiles",
                    "nodestats",
                    "ordered_traces",
                    "splining",
                    "topostats_file_version",
                ]
                for key in old_keys_to_remove:
                    try:
                        raw_data.pop(key)
                    except KeyError:
                        continue
            return raw_data
        except FileNotFoundError:
            LOGGER.error(f"File Not Found : {self.img_path}")
            raise

    def load_asd(self) -> tuple[npt.NDArray, float]:
        """
        Extract image and pixel to nm scaling from .asd files.

        Returns
        -------
        tuple[npt.NDArray, float]
            A tuple containing the image and its pixel to nanometre scaling value.
        """
        try:
            frames: np.ndarray
            pixel_to_nm_scaling: float
            _: dict
            frames, pixel_to_nm_scaling, _ = asd.load_asd(file_path=self.img_path, channel=self.channel)
            LOGGER.debug(f"[{self.filename}] : Loaded image from : {self.img_path}")
        except FileNotFoundError:
            LOGGER.error(f"File not found. Path: {self.img_path}")
            raise

        return (frames, pixel_to_nm_scaling)

    def load_ibw(self) -> tuple[npt.NDArray, float]:
        """
        Load image from Asylum Research (Igor) .ibw files.

        Returns
        -------
        tuple[npt.NDArray, float]
            A tuple containing the image and its pixel to nanometre scaling value.
        """
        try:
            LOGGER.debug(f"Loading image from : {self.img_path}")
            return ibw.load_ibw(file_path=self.img_path, channel=self.channel)
        except FileNotFoundError:
            LOGGER.error(f"File not found : {self.img_path}")
            raise

    def load_jpk(self) -> tuple[npt.NDArray, float]:
        """
        Load image from JPK Instruments .jpk files.

        Returns
        -------
        tuple[npt.NDArray, float]
            A tuple containing the image and its pixel to nanometre scaling value.
        """
        try:
            return jpk.load_jpk(file_path=self.img_path, channel=self.channel)
        except FileNotFoundError:
            LOGGER.error(f"[{self.filename}] File not found : {self.img_path}")
            raise

    def load_gwy(self) -> tuple[npt.NDArray, float]:
        """
        Extract image and pixel to nm scaling from the Gwyddion .gwy file.

        Returns
        -------
        tuple[npt.NDArray, float]
            A tuple containing the image and its pixel to nanometre scaling value.
        """
        LOGGER.debug(f"Loading image from : {self.img_path}")
        try:
            return gwy.load_gwy(file_path=self.img_path, channel=self.channel)
        except FileNotFoundError:
            LOGGER.error(f"File not found : {self.img_path}")
            raise

    def load_top(self) -> tuple[npt.NDArray, float]:
        """
        Extract image and pixel to nm scaling from the WsXM .top file.

        Returns
        -------
        tuple[npt.NDArray, float]
            A tuple containing the image and its pixel to nanometre scaling value.
        """
        LOGGER.debug(f"Loading image from : {self.img_path}")
        try:
            return top.load_top(file_path=self.img_path)
        except FileNotFoundError:
            LOGGER.error(f"File not found : {self.img_path}")
            raise

    def load_stp(self) -> tuple[npt.NDArray, float]:
        """
        Extract image and pixel to nm scaling from the WsXM .stp file.

        Returns
        -------
        tuple[npt.NDArray, float]
            A tuple containing the image and its pixel to nanometre scaling value.
        """
        LOGGER.debug(f"Loading image from : {self.img_path}")
        try:
            return stp.load_stp(file_path=self.img_path)
        except FileNotFoundError:
            LOGGER.error(f"File not found : {self.img_path}")
            raise

    def get_data(self) -> None:  # noqa: C901  # pylint: disable=too-many-branches
        """Extract image, filepath and pixel to nm scaling value, and append these to the img_dic object."""
        suffix_to_loader = {
            ".spm": self.load_spm,
            ".jpk": self.load_jpk,
            ".jpk-qi-image": self.load_jpk,
            ".ibw": self.load_ibw,
            ".gwy": self.load_gwy,
            ".topostats": self.load_topostats,
            ".asd": self.load_asd,
            ".stp": self.load_stp,
            ".top": self.load_top,
        }
        for img_path in self.img_paths:
            if Path(img_path).is_dir():
                continue
            self.img_path = Path(img_path)
            self.filename = str(Path(img_path).parts[-1])
            suffix = Path(img_path).suffix
            LOGGER.info(f"Extracting image from {self.img_path}")
            LOGGER.debug(f"File extension : {suffix}")

            OLD_BRUKER_RE = re.compile(r"\.\d+$")

            # Check that the file extension is supported
            if suffix in suffix_to_loader:
                data = None
                try:
                    if suffix == ".topostats":
                        data = suffix_to_loader[suffix]()
                        self.image = data["image_original"]
                        self.pixel_to_nm_scaling = data["pixel_to_nm_scaling"]
                    else:
                        self.image, self.pixel_to_nm_scaling = suffix_to_loader[suffix]()
                except Exception as e:
                    if "Channel" in str(e) and "not found" in str(e):
                        LOGGER.warning(e)  # log the specific error message
                        LOGGER.warning(f"[{self.filename}] Channel {self.channel} not found, skipping image.")
                    else:
                        raise
                else:
                    if suffix == ".asd":
                        for index, frame in enumerate(self.image):
                            self._check_image_size_and_add_to_dict(image=frame, filename=f"{self.filename}_{index}")
                    # If we are loading a .topostats we don't check image size, but instead update config, version and
                    # img_path to the current values and add directly to dictionary
                    elif suffix == ".topostats":
                        data["img_path"] = self.img_path.parent / self.filename
                        data["config"] = self.config
                        data["topostats_version"] = __version__
                        self.img_dict[self.filename] = dict_to_topostats(dictionary=data)

                    # Otherwise check the size and add image to dictionary
                    else:
                        self._check_image_size_and_add_to_dict(image=self.image, filename=self.filename)
            elif OLD_BRUKER_RE.match(suffix):
                # This is an old Bruker file, treat as normal.
                LOGGER.debug(f"Old Bruker file detected, treating as {suffix_to_loader['.spm'].__name__}")
                try:
                    self.image, self.pixel_to_nm_scaling = self.load_spm()
                except FileNotFoundError:
                    LOGGER.error(f"File not found : {self.img_path}")
                    raise
                self._check_image_size_and_add_to_dict(image=self.image, filename=self.filename)
            else:
                raise ValueError(f"File type {suffix} not yet supported. Please make an issue at \
                https://github.com/AFM-SPM/TopoStats/issues, or email topostats@sheffield.ac.uk to request support for \
                this file type.")

    def _check_image_size_and_add_to_dict(self, image: npt.NDArray, filename: str) -> None:
        """
        Check the image is above a minimum size in both dimensions.

        Images that do not meet the minimum size are not included for processing.

        Parameters
        ----------
        image : npt.NDArray
            An array of the extracted AFM image.
        filename : str
            The name of the file.
        """
        if image.shape[0] < self.MINIMUM_IMAGE_SIZE or image.shape[1] < self.MINIMUM_IMAGE_SIZE:
            LOGGER.warning(f"[{filename}] Skipping, image too small: {image.shape}")
        else:
            self.add_to_dict(image=image, filename=filename)
            LOGGER.debug(f"[{filename}] Image added to processing.")

    def add_to_dict(self, image: npt.NDArray, filename: str) -> None:
        """
        Add an image and metadata to the img_dict dictionary under the key filename.

        Adds the image and associated metadata such as any grain masks, and pixel to nanometere
        scaling factor to the img_dict dictionary which is used as a place to store the image
        information for processing.

        Parameters
        ----------
        image : npt.NDArray
            An array of the extracted AFM image.
        filename : str
            The name of the file.
        """
        self.img_dict[filename] = TopoStats(
            grain_crops=None,
            filename=filename,
            pixel_to_nm_scaling=self.pixel_to_nm_scaling,
            topostats_version=__version__,
            img_path=self.img_path.with_name(filename),
            image=None,
            image_original=image,
            config=deepcopy(self.config),
        )


def dict_to_hdf5(  # noqa: C901 # pylint: disable=too-many-statements
    open_hdf5_file: h5py.File, group_path: str, dictionary: dict
) -> None:
    """
    Recursively save a dictionary to an open hdf5 file.

    Parameters
    ----------
    open_hdf5_file : h5py.File
        An open hdf5 file object.
    group_path : str
        The path to the group in the hdf5 file to start saving data from.
    dictionary : dict
        A dictionary of the data to save.
    """
    for key, item in dictionary.items():
        LOGGER.debug(f"Saving key: {key}")
        if item is None:
            LOGGER.debug(f"Item '{key}' is None. Skipping.")
        # Make sure the key is a string
        key = str(key)

        # Check if the item is a known datatype
        if isinstance(
            item,
            (
                list
                | str
                | int
                | float
                | np.ndarray
                | Path
                | dict
                | DisorderedTrace
                | GrainCrop
                | MatchedBranch
                | Molecule
                | Node
                | OrderedTrace
            ),
        ):  # noqa: UP038
            # Exclude the plotting dictionary configuration, its excessive and users rarely tinker with it
            if key == "plot_dict":
                continue
            # Lists need to be converted to numpy arrays
            if isinstance(item, list):
                LOGGER.debug(f"[dict_to_hdf5] {key} is of type : {type(item)}")
                if any(x is None for x in item):
                    item = np.array([np.nan if x is None else x for x in item])
                open_hdf5_file.create_dataset(name=group_path + key, data=item, compression="gzip")
            # Strings need to be encoded to bytes, but can not be compressed
            elif isinstance(item, str):
                LOGGER.debug(f"[dict_to_hdf5] {key} is of type : {type(item)}")
                open_hdf5_file.create_dataset(name=group_path + key, data=item.encode("utf8"))
            # Integers, floats are added directly to the hdf5 file
            elif isinstance(item, (int, float)):  # noqa: UP038
                LOGGER.debug(f"[dict_to_hdf5] {key} is of type : {type(item)}")
                open_hdf5_file.create_dataset(name=group_path + key, data=item)
            # Numpy arrays can be compressed added directly to the hdf5 file
            elif isinstance(item, np.ndarray):  # noqa: UP038
                LOGGER.debug(f"[dict_to_hdf5] {key} is of type : {type(item)}")
                open_hdf5_file.create_dataset(name=group_path + key, data=item, compression="gzip")
            # Path objects need to be converted to strings which can not be compressed
            elif isinstance(item, Path):
                LOGGER.debug(f"[dict_to_hdf5] {key} is of type : {type(item)}")
                open_hdf5_file.create_dataset(name=group_path + key, data=str(item).encode("utf8"))
            # Dictionaries need to be recursively saved
            elif isinstance(item, dict):  # a sub-dictionary, so we need to recurse
                LOGGER.debug(f"[dict_to_hdf5] {key} is of type : {type(item)}")
                dict_to_hdf5(open_hdf5_file, group_path + key + "/", item)
            # All classes defined in the classes submodule must be recursively saved
            elif isinstance(item, (GrainCrop | OrderedTrace | Node | DisorderedTrace | MatchedBranch | Molecule)):
                LOGGER.debug(f"[dict_to_hdf5] {key} is of type : {type(item)}")
                dict_to_hdf5(open_hdf5_file, group_path + key + "/", convert_to_dict(item))

        else:  # attempt to save an item that is not a numpy array or a dictionary
            try:
                open_hdf5_file[group_path + key] = item
            except Exception as e:
                LOGGER.debug(f"Cannot save key '{key}' to HDF5. Item type: {type(item)}. Skipping. {e}")


def hdf5_to_dict(open_hdf5_file: h5py.File, group_path: str) -> dict:
    """
    Read a dictionary from an open hdf5 file.

    Parameters
    ----------
    open_hdf5_file : h5py.File
        An open hdf5 file object.
    group_path : str
        The path to the group in the hdf5 file to start reading data from.

    Returns
    -------
    dict
        A dictionary of the hdf5 file data.
    """
    data_dict = {}
    for key, item in open_hdf5_file[group_path].items():
        LOGGER.debug(f"Loading hdf5 key: {key}")
        if isinstance(item, h5py.Group):
            LOGGER.debug(f" {key} is a group")
            data_dict[key] = hdf5_to_dict(open_hdf5_file, group_path + key + "/")
        # Decode byte strings to utf-8. The data type "O" is a byte string.
        elif isinstance(item, h5py.Dataset) and item.dtype == "O":
            LOGGER.debug(f" {key} is a byte string")
            data_dict[key] = item[()].decode("utf-8")
            LOGGER.debug(f" {key} type: {type(data_dict[key])}")
        else:
            LOGGER.debug(f" {key} is other type of dataset")
            data_dict[key] = item[()]
            LOGGER.debug(f" {key} type: {type(data_dict[key])}")
    return data_dict


def save_topostats_file(output_dir: Path, topostats_object: TopoStats, topostats_version: str = __release__) -> None:
    """
    Save ''TopoStats'' object to a ''.topostats'' (hdf5 format) file.

    Parameters
    ----------
    output_dir : Path
        Directory to save the .topostats file in.
    topostats_object : dict
        Dictionary of the topostats data to save. Must include a flattened image and pixel to nanometre scaling
        factor. May also include grain masks.
    topostats_version : str
        Version to save as, defaults to ''__release__''.
    """
    LOGGER.info(f"[{topostats_object.filename}] : Saving image to .topostats file")
    if ".topostats" not in topostats_object.filename:
        # Remove everything after the last period in filename)
        filename = topostats_object.filename.rsplit(".", 1)[0]
        save_file_path = Path(output_dir) / f"{filename}.topostats"
    else:
        save_file_path = Path(output_dir) / topostats_object.filename
    with h5py.File(save_file_path, "w") as f:
        # It may be possible for topostats_object["image"] to be None.
        # Make sure that this is not the case.
        if isinstance(topostats_object, dict) and topostats_object["image"] is not None:
            # Recursively save the topostats object dictionary to the .topostats file
            if parse_version(topostats_version) < parse_version("2.4"):
                topostats_object["topostats_file_version"] = topostats_version
                dict_to_hdf5(open_hdf5_file=f, group_path="/", dictionary=topostats_object)
        elif isinstance(topostats_object, TopoStats):
            topostats_object.topostats_version = topostats_version
            dict_to_hdf5(open_hdf5_file=f, group_path="/", dictionary=convert_to_dict(topostats_object))

        else:
            raise ValueError(
                "TopoStats object dictionary does not contain an 'image'. "
                "TopoStats objects must be saved with a flattened image."
            )


def dict_to_topostats(  # noqa: C901 # pylint: disable=too-many-locals,too-many-nested-blocks
    dictionary: dict[str, Any],
) -> TopoStats:
    """
    Convert a dictionary, typically loaded from HDF5 ``.topostats`` file, to TopoStats object.

    Parameters
    ----------
    dictionary : dict[str, Any]
        Dictionary of TopoStats data. This will typically have been loaded from the HDF5 ``.topostats`` file using
        `AFMReader <https://afm-spm.github.io/AFMReader>`__.

    Returns
    -------
    TopoStats
        A TopoStat object.
    """
    if "grain_crops" in dictionary.keys() and dictionary["grain_crops"] is not None:
        grain_crops = {}
        for grain, crop in dictionary["grain_crops"].items():
            image = crop["image"] if "image" in crop.keys() else None
            mask = crop["mask"] if "mask" in crop.keys() else None
            padding = int(crop["padding"]) if "padding" in crop.keys() else None
            bbox = crop["bbox"] if "bbox" in crop.keys() else None
            pixel_to_nm_scaling = crop["pixel_to_nm_scaling"] if "pixel_to_nm_scaling" in crop.keys() else None
            filename = crop["filename"] if "filename" in crop.keys() else None
            skeleton = crop["skeleton"] if "skeleton" in crop.keys() else None
            height_profiles = crop["height_profiles"] if "height_profiles" in crop.keys() else None
            stats = crop["stats"] if "stats" in crop.keys() else None
            disordered_trace = (
                DisorderedTrace(**crop["disordered_trace"]) if "disordered_trace" in crop.keys() else None
            )
            if "nodes" in crop.keys() and crop["nodes"] is not None:
                nodes = {}
                for node, node_data in crop["nodes"].items():
                    nodes[node] = Node(**node_data)
            else:
                nodes = None
            ordered_trace = OrderedTrace(**crop["ordered_trace"]) if "ordered_trace" in crop.keys() else None
            threshold_method = crop["threshold_method"] if "threshold_method" in crop.keys() else None
            thresholds = crop["thresholds"] if "thresholds" in crop.keys() else None
            grain_crops[grain] = GrainCrop(
                image=image,
                mask=mask,
                padding=padding,
                bbox=bbox,
                pixel_to_nm_scaling=pixel_to_nm_scaling,
                filename=filename,
                skeleton=skeleton,
                height_profiles=height_profiles,
                stats=stats,
                disordered_trace=disordered_trace,
                nodes=nodes,
                ordered_trace=ordered_trace,
                threshold_method=threshold_method,
                thresholds=thresholds,
            )
    else:
        grain_crops = None
    full_mask_tensor = dictionary["full_mask_tensor"] if "full_mask_tensor" in dictionary.keys() else None
    config = dictionary["config"] if "config" in dictionary.keys() else None
    # We have to handle old (topostats_file_version) and new (topostats_version) keys and make sure they are str
    topostats_version = (
        str(dictionary["topostats_file_version"])
        if "topostats_file_version" in dictionary.keys()
        else str(dictionary["topostats_version"])
    )
    image_name = Path(dictionary["filename"]).stem
    return TopoStats(
        grain_crops=grain_crops,
        filename=dictionary["filename"],
        image_name=image_name,
        pixel_to_nm_scaling=dictionary["pixel_to_nm_scaling"],
        img_path=dictionary["img_path"],
        image=dictionary["image"],
        image_original=dictionary["image_original"],
        topostats_version=str(topostats_version),
        full_mask_tensor=full_mask_tensor,
        config=config,
    )


def save_pkl(outfile: Path, to_pkl: dict) -> None:
    """
    Pickle objects for working with later.

    Parameters
    ----------
    outfile : Path
        Path and filename to save pickle to.
    to_pkl : dict
        Object to be pickled.
    """
    with outfile.open(mode="wb", encoding=None) as f:
        pkl.dump(to_pkl, f)


def load_pkl(infile: Path) -> Any:
    """
    Load data from a pickle.

    Parameters
    ----------
    infile : Path
        Path to a valid pickle.

    Returns
    -------
    dict:
        Dictionary of generated images.

    Examples
    --------
    from pathlib import Path

    from topostats.io import load_plots

    pkl_path = "output/distribution_plots.pkl"
    my_plots = load_pkl(pkl_path)
    # Show the type of my_plots which is a dictionary of nested dictionaries
    type(my_plots)
    # Show the keys are various levels of nesting.
    my_plots.keys()
    my_plots["area"].keys()
    my_plots["area"]["dist"].keys()
    # Get the figure and axis object for a given metrics distribution plot
    figure, axis = my_plots["area"]["dist"].values()
    # Get the figure and axis object for a given metrics violin plot
    figure, axis = my_plots["area"]["violin"].values()
    """
    with infile.open("rb", encoding=None) as f:
        return pkl.load(f)  # noqa: S301


def dict_to_json(data: dict, output_dir: str | Path, filename: str | Path, indent: int = 4) -> None:
    """
    Write a dictionary to a JSON file at the specified location with the given name.

    NB : The `NumpyEncoder` class is used as the default encoder to ensure Numpy dtypes are written as strings (they are
         not serialisable to JSON using the default JSONEncoder).

    Parameters
    ----------
    data : dict
        Data as a dictionary that is to be written to file.
    output_dir : str | Path
        Directory the file is to be written to.
    filename : str | Path
        Name of output file.
    indent : int
        Spaces to indent JSON with, default is 4.
    """
    output_file = output_dir / filename
    with output_file.open("w") as f:
        json.dump(data, f, indent=indent, cls=NumpyEncoder)


def write_csv(
    df: pd.DataFrame,
    dataset: str,
    names: list[str] | None,
    index: list[str],
    output_dir: str | Path,
    base_dir: str | Path,
) -> pd.DataFrame:
    """
    Write summary statistics files to CSV.

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe to write to CSV.
    dataset : str
        Type of dataframe, valid values are ``grain_stats``, ``matched_branch_stats``, ``branch_statistics`` or
        ``mol_stats``.
    names : list[str], optional
        List of names to rename current index with.
    index : list[str]
        List of columns to set index to.
    output_dir : str | Path
        Output directory.
    base_dir : str | Path
        Base directory.

    Returns
    -------
    pd.DataFrame
        Pandas dataframe with index renamed and reset.
    """
    if dataset not in {"grain_stats", "matched_branch_stats", "branch_statistics", "mol_stats"}:
        raise ValueError(
            f"Invalid dataset {dataset}, valid datasets for writing to CSV are "
            "'grain_stats', 'matched_branch_stats', 'branch_statistics', 'mol_stats'."
        )
    dataset_to_csv = {
        "grain_stats": "grain_statistics.csv",
        "matched_branch_stats": "matched_branch_statistics.csv",
        "branch_statistics": "branch_statistics.csv",
        "mol_stats": "molecule_statistics.csv",
    }
    # Set index names
    df.index.set_names(names, inplace=True)
    # Reset index
    if dataset == "mol_stats":
        df.reset_index(drop=True, inplace=True)
    else:
        df.reset_index(inplace=True)
    # Set index to required columns
    df.set_index(index, inplace=True)
    # Write to CSV
    df.to_csv(Path(output_dir) / dataset_to_csv[dataset], index=True)
    # Write statistics on per-folder basis
    save_image_grainstats(output_dir=output_dir, base_dir=base_dir, all_stats_df=df, stats_filename=dataset)
    # Return dataframe with index reset
    df.reset_index(inplace=True)
    return df


def extract_height_profiles(
    topostats_object_all: dict[str, TopoStats],
    output_dir: str | Path,
    filename: str = "height_profiles.json",
) -> None:
    """
    Write height profiles from all grains across all processed images to JSON.

    Parameters
    ----------
    topostats_object_all : dict[str, TopoStats]
        Dictionary of processed TopoStats objects.
    output_dir : str | Path
        Path to save JSON to.
    filename : str
        Name of output file, default is ``height_profile.json``.
    """
    height_profile_all = {}
    for image, topostats_object in topostats_object_all.items():
        height_profile_all[image] = {}
        for grain_number, grain_crop in topostats_object.grain_crops.items():
            height_profile_all[image][grain_number] = grain_crop.height_profiles
    dict_to_json(data=height_profile_all, output_dir=output_dir, filename=filename)
    LOGGER.info(f"Saved all height profiles to {output_dir}/{filename}.")



"""Image artefact correction functions that interpolates values filling the space of any detected scars."""

import logging

import numpy as np
import numpy.typing as npt

from topostats.logs.logs import LOGGER_NAME

LOGGER = logging.getLogger(LOGGER_NAME)

# pylint: disable=too-many-arguments
# pylint: disable=too-many-instance-attributes


def _mark_if_positive_scar(
    row_col: tuple,
    stddev: float,
    img: npt.NDArray,
    marked: npt.NDArray,
    threshold_low: float,
    max_scar_width: int,
) -> None:
    """
    Mark scars as positive (i.e. a ridge rather than a dip).

    Determine if the points below and including the pixel at the specified row and column are a positive scar (a
    ridge rather than a dip). If they are, mark them in the marked 2d npt.NDArray. Note that this only detects positive
    scars.

    Parameters
    ----------
    row_col : tuple
        A tuple containing the row and column indices of the pixel for the top of the potential scar. Note that
        the first value is the row index, and the second is the column index.
    stddev : float
        The standard deviation, or the root-mean-square value for the image.
    img : npt.NDArray
        A 2-D image of the data to remove scars from.
    marked : np.ndarry
        A 2-D image of pixels that stores the positions of scars marked for removal. The value of the pixel is a
        floating point value that represents how strongly the algorithm considers it to be a scar.
        This may or may not already contain non-zero values given that previous iterations of scar removal
        may have been performed.
    threshold_low : float
        A value that when multiplied with the standard deviation, acts as a threshold to determine if an increase
        or decrease in height might constitute the top or bottom of a scar.
    max_scar_width : int
        A value that dictates the maximum width that a scar can be. Note that this does not mean horizontal width,
        rather vertical, this is because we consider scars to be laying flat, horizontally, so their width is
        vertical and their length is horizontal.
    """
    # Unpack row, col
    row = row_col[0]
    col = row_col[1]

    # If sharp enough rise
    min_scar_value = img[row + 1, col]
    max_border_value = img[row, col]
    if min_scar_value - max_border_value > threshold_low * stddev:
        # Possibly in a scar
        for k in range(1, max_scar_width + 1):
            if row + k + 1 >= img.shape[0]:
                # Bottom of image, break
                LOGGER.debug("Bottom of image.")
                break
            min_scar_value = min(min_scar_value, img[row + k, col])
            max_border_value = max(img[row, col], img[row + k + 1, col])
            # Check if scar ended
            if min_scar_value - max_border_value > threshold_low * stddev:
                while k:
                    val = (img[row + k, col] - max_border_value) / stddev
                    marked[row + k, col] = val
                    k -= 1


def _mark_if_negative_scar(
    row_col: tuple,
    stddev: float,
    img: npt.NDArray,
    marked: npt.NDArray,
    threshold_low: float,
    max_scar_width: int,
) -> None:
    """
    Mark scars as negative (i.e. a dip rather than a ridge).

    Determine if the points below and including the pixel at the specified row and column are a negative scar (a
    dip rather than a ridge). If they are, mark them in the marked 2d npt.NDArray. Note that this only detects negative
    scars.

    Parameters
    ----------
    row_col : tuple
        A tuple containing the row and column indices of the pixel for the top of the potential scar. Note that
        the first value is the row index, and the second is the column index.
    stddev : float
        The standard deviation, or the root-mean-square value for the image.
    img : npt.NDArray
        A 2-D image of the data to remove scars from.
    marked : np.ndarry
        A 2-D image of pixels that stores the positions of scars marked for removal. The value of the pixel is a
        floating point value that represents how strongly the algorithm considers it to be a scar.
        This may or may not already contain non-zero values given that previous iterations of scar removal
        may have been performed.
    threshold_low : float
        A value that when multiplied with the standard deviation, acts as a threshold to determine if an increase
        or decrease in height might constitute the top or bottom of a scar.
    max_scar_width : int
        A value that dictates the maximum width that a scar can be. Note that this does not mean horizontal width,
        rather vertical, this is because we consider scars to be laying flat, horizontally, so their width is
        vertical and their length is horizontal.
    """
    # Unpack row, col
    row = row_col[0]
    col = row_col[1]

    # If sharp enough dip
    min_scar_value = img[row + 1, col]
    max_border_value = img[row, col]
    if min_scar_value - max_border_value < -threshold_low * stddev:
        # Possibly in a scar
        for k in range(1, max_scar_width + 1):
            if row + k + 1 >= img.shape[0]:
                # Bottom of image, break
                LOGGER.debug("Bottom of image.")
                break
            min_scar_value = max(min_scar_value, img[row + k, col])
            max_border_value = min(img[row, col], img[row + k + 1, col])
            # Check if scar ended
            if min_scar_value - max_border_value < -threshold_low * stddev:
                while k:
                    val = (max_border_value - img[row + k, col]) / stddev
                    marked[row + k, col] = val
                    k -= 1


def _spread_scars(
    marked: npt.NDArray,
    threshold_low: float,
    threshold_high: float,
) -> None:
    """
    Spread high-marked pixels into adjacent low-marked pixels.

    This is a smudging function that attempts to catch any pixels that are parts of scars that might not have been
    extreme enough to get marked above the high_threshold. Any remaining marked pixels below high_threshold are
    considered not to be scars and are removed from the mask.

    Parameters
    ----------
    marked : npt.NDArray
        A 2-D image of pixels that stores the positions of scars marked for removal. The value of the pixel is a
        floating point value that represents how strongly the algorithm considers it to be a scar.
        This may or may not already contain non-zero values given that previous iterations of scar removal
        may have been performed.
    threshold_low : float
        A floating point value, that when multiplied by the standard deviation of the image, acts as a
        threshold for sharp inclines or descents in pixel values and thus marks potential scars.
        A lower value will make the algorithm
        more sensitive, and a higher value will make it less sensitive.
    threshold_high : float
        A floating point value that is used similarly to threshold_low, however sharp inclines or descents
        that result in values in the mask higher than this threshold are automatically considered scars.
    """
    # Spread scars that have close to threshold edge-points
    for row in range(marked.shape[0]):
        # Spread right
        for col in range(1, marked.shape[1]):
            if marked[row, col] >= threshold_low and marked[row, col - 1] >= threshold_high:
                marked[row, col] = threshold_high

        # Spread left
        for col in range(marked.shape[1] - 1, 0, -1):
            if marked[row, col - 1] >= threshold_low and marked[row, col] >= threshold_high:
                marked[row, col - 1] = threshold_high


def _remove_short_scars(marked: npt.NDArray, threshold_high: float, min_scar_length: int) -> None:
    """
    Remove scars that are too short (horizontally), based on the minimum length.

    Parameters
    ----------
    marked : npt.NDArray
        A 2-D image of pixels that stores the positions of scars marked for removal. The value of the pixel is a
        floating point value that represents how strongly the algorithm considers it to be a scar.
        This may or may not already contain non-zero values given that previous iterations of scar removal
        may have been performed.

    threshold_high : float
        A floating point value that is used similarly to threshold_low, however sharp inclines or descents
        that result in values in the mask higher than this threshold are automatically considered scars.

    min_scar_length : int
        A value that dictates the maximum width that a scar can be. Note that this does not mean horizontal width,
        rather vertical, this is because we consider scars to be laying flat, horizontally, so their width is
        vertical and their length is horizontal.
    """
    # Remove too-short scars
    for row in range(marked.shape[0]):
        k = 0
        for col in range(marked.shape[1]):
            # If greater than threshold, set value to true
            if marked[row, col] >= threshold_high:
                marked[row, col] = 1.0
                k += 1
                continue
            # If not greater than threshold,
            # either we reached the end of the scar,
            # or haven't found one.
            # Check if too short. If so, remove.
            if k and k < min_scar_length:
                while k:
                    marked[row, col - k] = 0.0
                    k -= 1
            # Long enough, just not bright anymore.
            # Reached the end of the scar that is long enough. Stop and reset.
            marked[row, col] = 0.0
            k = 0
        # Reached the end of the line, so check current scar's
        # length to see if too short and should be deleted.
        if k and k < min_scar_length:
            while k:
                marked[row, col - k] = 0.0
                k -= 1


def _mark_scars(
    img: npt.NDArray,
    direction: str,
    threshold_low: float,
    threshold_high: float,
    max_scar_width: int,
    min_scar_length: int,
) -> npt.NDArray:
    """
    Mark scars within an image, returning a boolean 2D npt.NDArray of pixels that have been detected as scars.

    Parameters
    ----------
    img : npt.NDArray
        A 2-D image of the data to detect scars in.
    direction : str
        Options : 'positive', 'negative'. The direction of scars to detect. For example, to detect scars
        that lie above the data, select 'positive'.
    threshold_low : float
        A floating point value, that when multiplied by the standard deviation of the image, acts as a
        threshold for sharp inclines or descents in pixel values and thus marks potential scars.
        A lower value will make the algorithm
        more sensitive, and a higher value will make it less sensitive.
    threshold_high : float
        A floating point value that is used similarly to threshold_low, however sharp inclines or descents
        that result in values in the mask higher than this threshold are automatically considered scars.
    max_scar_width : int
        An integer that restricts the algorithm to only mark scars that are as thin or thinner than this width.
        This is important for ensuring that legitimate grain data is not detected as scarred data.
        Note that width here is vertical, as scars are thin, horizontal features.
    min_scar_length : int
        An integer that restricts the algorithm to only mark scars that are as long or longer than this length.
        This is important for ensuring that noise or legitimate but sharp datapoints do not get detected as scars.
        Note that length here is horizontal, as scars are thin, horizontal features.

    Returns
    -------
    marked: npt.NDArray
        Returns a 2-D image of the same shape as the data image, where each pixel's value represents
        a metric for how strongly that pixel is considered to be a scar. Higher values mean more likely
        to be a scar.
    """
    image = np.copy(img)

    stddev = np.std(image)
    marked = np.zeros(image.shape)

    for row in range(image.shape[0] - 1):
        for col in range(image.shape[1]):
            if direction == "positive":
                _mark_if_positive_scar(
                    row_col=(row, col),
                    stddev=stddev,
                    img=image,
                    marked=marked,
                    threshold_low=threshold_low,
                    max_scar_width=max_scar_width,
                )

            elif direction == "negative":
                _mark_if_negative_scar(
                    row_col=(row, col),
                    stddev=stddev,
                    img=image,
                    marked=marked,
                    threshold_low=threshold_low,
                    max_scar_width=max_scar_width,
                )

            else:
                raise ValueError(f"direction {direction} invalid.")

    _spread_scars(marked=marked, threshold_low=threshold_low, threshold_high=threshold_high)

    _remove_short_scars(marked=marked, threshold_high=threshold_high, min_scar_length=min_scar_length)

    return marked


def _remove_marked_scars(img: npt.NDArray, scar_mask: npt.NDArray) -> None:
    """
    Interpolate values covered by marked scars.

    Takes an image, and a marked scar boolean mask for that image. Returns the image where the marked scars are replaced
    by interpolated values.

    Parameters
    ----------
    img : npt.NDArray
        A 2-D image of the data to remove scars from.
    scar_mask : npt.NDArray
        A boolean image of pixels that determine which values are flagged as scars and therefore should
        be interpolated over in the original data image.
    """
    for row, col in np.ndindex(img.shape):
        if scar_mask[row, col] == 1.0:
            # Determine how wide the scar is by incrementing scar_width until either the bottom
            # of the image is reached, or a non-marked pixel is encountered.
            scar_width = 1
            while row + scar_width < img.shape[0] and scar_mask[row + scar_width, col] == 1.0:
                scar_width += 1

            above = img[row - 1, col]
            below = img[row + scar_width, col]
            k = scar_width
            while k:
                # Linearly interpolate
                interp_val = (k / (scar_width + 1)) * below + (1 - (k / (scar_width + 1))) * above
                img[row + k - 1, col] = interp_val
                scar_mask[row + k - 1, col] = 0.0
                k -= 1
                LOGGER.debug("Scar removed")


def remove_scars(
    img: npt.NDArray,
    filename: str,
    removal_iterations: int = 2,
    threshold_low: float = 0.250,
    threshold_high: float = 0.666,
    max_scar_width: int = 4,
    min_scar_length: int = 16,
):
    """
    Remove scars from an image.

    Scars are long, typically 1-4 pixels wide streaks of high or low data in AFM images. They are a problem
    resulting from random errors in the AFM data collection process and are hard to avoid. This function
    detects and removes these artefacts by interpolating over them between the pixels above and below them.
    This method takes no parameters as it uses parameters already established as instance variables when the
    class was instantiated.

    Parameters
    ----------
    img : npt.NDArray
        A 2-D image to remove scars from.
    filename : str
        The filename (used for logging outputs only).
    removal_iterations : int
        The number of times the scar removal should run on the image.
        Running just once sometimes isn't enough to remove some of the
        more difficult to remove scars.
    threshold_low : float
        A value that when multiplied with the standard deviation, acts as a threshold to determine if an increase
        or decrease in height might constitute the top or bottom of a scar.
    threshold_high : float
        A floating point value that is used similarly to threshold_low, however sharp inclines or descents
        that result in values in the mask higher than this threshold are automatically considered scars.
    max_scar_width : int
        A value that dictates the maximum width that a scar can be. Note that this does not mean horizontal width,
        rather vertical, this is because we consider scars to be laying flat, horizontally, so their width is
        vertical and their length is horizontal.
    min_scar_length : int
        An integer that restricts the algorithm to only mark scars that are as long or longer than this length.
        This is important for ensuring that noise or legitimate but sharp datapoints do not get detected as scars.
        Note that length here is horizontal, as scars are thin, horizontal features.

    Returns
    -------
    self.img
        The original 2-D image with scars removed, unless the config has run set to False, in which case it
        will not remove the scars.
    """
    LOGGER.info(f"[{filename}] : Removing scars")

    first_marked_mask = None
    for i in range(removal_iterations):
        marked_positive = _mark_scars(
            img=img,
            direction="positive",
            threshold_low=threshold_low,
            threshold_high=threshold_high,
            max_scar_width=max_scar_width,
            min_scar_length=min_scar_length,
        )
        marked_negative = _mark_scars(
            img=img,
            direction="negative",
            threshold_low=threshold_low,
            threshold_high=threshold_high,
            max_scar_width=max_scar_width,
            min_scar_length=min_scar_length,
        )
        # Combine the upper and lower scar masks
        marked_both = np.bitwise_or(marked_positive.astype(bool), marked_negative.astype(bool))

        if i == 0:
            first_marked_mask = marked_both

        _remove_marked_scars(img, np.copy(marked_both))

        LOGGER.debug("Scars removed")

    return img, first_marked_mask



"""Define custom classes for TopoStats."""

from __future__ import annotations

import logging
import re
from pathlib import Path
from typing import Any

import numpy as np
import numpy.typing as npt
from pydantic import ConfigDict
from pydantic.dataclasses import dataclass

from topostats.logs.logs import LOGGER_NAME
from topostats.statistics import image_statistics
from topostats.utils import update_background_class

LOGGER = logging.getLogger(LOGGER_NAME)

# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-arguments
# pylint: disable=too-many-positional-arguments
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-lines
# F811 - redefined-while-unused : we disable this as we want to have getter/setter methods for each attribute
# ruff: noqa: F811


class GrainCrop:
    """
    Class for storing the crops of grains.

    Parameters
    ----------
    image : npt.NDArray[np.float32]
        2-D Numpy array of the cropped image.
    mask : npt.NDArray[np.bool_]
        3-D Numpy tensor of the cropped mask.
    padding : int
        Padding added to the bounding box of the grain during cropping.
    bbox : tuple[int, int, int, int]
        Bounding box of the crop including padding.
    pixel_to_nm_scaling : float
        Pixel to nanometre scaling factor for the crop.
    thresholds : list[float]
        Thresholds used to find the grain.
    filename : str
        Filename of the image from which the crop was taken.
    threshold : str
        Direction of the molecule from the threshold (above / below).
    skeleton : npt.NDArray[np.bool_]
        3-D Numpy tensor of the skeletonised mask.
    convolved_skeleton : npt.NDArray[np.int32] | None = None
        2-D Numpy array of the convolved skeleton.
    height_profiles : dict[int, dict[int, npt.NDArray[np.float32]]] | None
        Nested dictionary  height profiles.
    stats : dict[int, dict[int, Any]] | None
        Dictionary of grain statistics.
    disordered_trace : DisorderedTrace
        A disordered trace for the grain.
    nodes : dict[str, Node]
        Dictionary of grain nodes.
    ordered_trace : OrderedTrace
        An ordered trace for the grain.
    threshold_method : str
        Threshold method used to find grains.
    """

    def __init__(  # pylint: disable=too-many-locals
        self,
        image: npt.NDArray[np.float32],
        mask: npt.NDArray[np.bool_],
        padding: int,
        bbox: tuple[int, int, int, int],
        pixel_to_nm_scaling: float,
        thresholds: list[float],
        filename: str,
        threshold: str | None = None,
        skeleton: npt.NDArray[np.bool_] | None = None,
        convolved_skeleton: npt.NDArray[np.int32] | None = None,
        height_profiles: dict[int, dict[int, npt.NDArray[np.float32]]] | None = None,
        stats: dict[int, dict[int, Any]] | None = None,
        disordered_trace: DisorderedTrace | None = None,
        nodes: dict[str, Node] | None = None,
        ordered_trace: OrderedTrace | None = None,
        threshold_method: str | None = None,
    ):
        """
        Initialise the class.

        Parameters
        ----------
        image : npt.NDArray[np.float32]
            2-D Numpy array of the cropped image.
        mask : npt.NDArray[np.bool_]
            3-D Numpy tensor of the cropped mask.
        padding : int
            Padding added to the bounding box of the grain during cropping.
        bbox : tuple[int, int, int, int]
            Bounding box of the crop including padding.
        pixel_to_nm_scaling : float
            Pixel to nanometre scaling factor for the crop.
        thresholds : list[float]
            A list of thresholds used to identify the grain.
        filename : str
            Filename of the image from which the crop was taken.
        threshold : str
            Direction of the molecule from the threshold (above / below).
        skeleton : npt.NDArray[np.bool_]
            3-D Numpy tensor of the skeletonised mask.
        convolved_skeleton : npt.NDArray[np.int32] | None = None
            2-D Numpy array of the convolved skeleton.
        height_profiles : dict[int, dict[int, npt.NDArray[np.float32]]] | None
            3-D Numpy tensor of the height profiles.
        stats : dict[str, int | float] | None
            Dictionary of grain statistics.
        disordered_trace : DisorderedTrace
            A disordered trace for the current grain.
        nodes : dict[int, Node] | None
            Grain nodes.
        ordered_trace : OrderedTrace | None
            An ordered trace for the grain.
        threshold_method : str
            Threshold method used to find grains.
        """
        self.padding = padding
        self.image = image
        # This part of the constructor must go after padding since the setter
        # for mask requires the padding.
        self.mask = mask
        self.bbox = bbox
        self.pixel_to_nm_scaling = pixel_to_nm_scaling
        self.thresholds = thresholds
        self.filename = filename
        self.threshold: str | None = threshold
        self.height_profiles: dict[int, dict[int, npt.NDArray[np.float32]]] | None = height_profiles
        self.stats: dict[str, Any] | None = {} if stats is None else stats
        self.skeleton: npt.NDArray[np.bool_] | None = skeleton
        self.convolved_skeleton: npt.NDArray[np.int32] | None = convolved_skeleton
        self.disordered_trace: DisorderedTrace | None = disordered_trace
        self.nodes: dict[int, Node] | None = nodes
        self.ordered_trace: OrderedTrace | None = ordered_trace
        self.threshold_method: str | None = threshold_method

    def __eq__(self, other: object) -> bool:
        """
        Check if two GrainCrop objects are equal.

        Parameters
        ----------
        other : object
            Object to compare to.

        Returns
        -------
        bool
            True if the objects are equal, False otherwise.
        """
        if not isinstance(other, GrainCrop):
            return False
        return (
            np.array_equal(self.image, other.image)
            and np.array_equal(self.mask, other.mask)
            and self.padding == other.padding
            and self.bbox == other.bbox
            and self.pixel_to_nm_scaling == other.pixel_to_nm_scaling
            and self.thresholds == other.thresholds
            and self.filename == other.filename
            and self.threshold == other.threshold
            and self.height_profiles == other.height_profiles
            and self.stats == other.stats
            and np.array_equal(self.skeleton, other.skeleton)
            and self.convolved_skeleton == other.convolved_skeleton
            and self.disordered_trace == other.disordered_trace
            and self.nodes == other.nodes
            and self.ordered_trace == other.ordered_trace
            and self.threshold_method == other.threshold_method
        )

    def __str__(self) -> str:
        """
        Representation function for useful statistics for the class.

        Returns
        -------
        str
            Set of formatted statistics on matched branches.
        """
        return (
            f"filename : {self.filename}\n"
            f"image shape : {self.image.shape}\n"
            f"skeleton shape : {self.skeleton.shape}\n"
            f"mask shape : {self.mask.shape}\n"
            f"padding : {self.padding}\n"
            f"thresholds : {self.thresholds}\n"
            f"threshold method : {self.threshold_method}\n"
            f"bounding box coords : {self.bbox}\n"
            f"pixel to nm scaling : {self.pixel_to_nm_scaling}\n"
            f"number of nodes : {len(self.nodes)}"
        )

    @property
    def image(self) -> npt.NDArray[float]:
        """
        Getter for the ``image`` attribute.

        Returns
        -------
        npt.NDArray
            Numpy array of the image.
        """
        return self._image

    @image.setter
    def image(self, value: npt.NDArray[float]):
        """
        Setter for the ``image`` attribute.

        Parameters
        ----------
        value : npt.NDArray
            Numpy array of the image.

        Raises
        ------
        ValueError
            If the image is not square.
        """
        if value.shape[0] != value.shape[1]:
            raise ValueError(f"Image is not square: {value.shape}")
        self._image = value

    @property
    def mask(self) -> npt.NDArray[np.bool_]:
        """
        Getter for the ``mask`` attribute.

        Returns
        -------
        npt.NDArray[np.bool_]
            Numpy array of the mask.
        """
        return self._mask

    @mask.setter
    def mask(self, value: npt.NDArray[np.bool_]) -> None:
        """
        Setter for the ``mask`` attribute.

        Parameters
        ----------
        value : npt.NDArray
            Numpy array of the ``mask`` attribute.

        Raises
        ------
        ValueError
            If the mask dimensions do not match the image.
        """
        if value.shape[0] != self.image.shape[0] or value.shape[1] != self.image.shape[1]:
            raise ValueError(f"Mask dimensions do not match image: {value.shape} vs {self.image.shape}")
        # Ensure that the padding region is blank, set it to be blank if not
        try:
            for class_index in range(1, value.shape[2]):
                class_mask = value[:, :, class_index]

                padded_region_top = class_mask[: self.padding, :]
                padded_region_bottom = class_mask[-self.padding :, :]
                padded_region_left = class_mask[:, : self.padding]
                padded_region_right = class_mask[:, -self.padding :]
                if (
                    np.any(padded_region_top)
                    or np.any(padded_region_bottom)
                    or np.any(padded_region_left)
                    or np.any(padded_region_right)
                ):
                    LOGGER.warning("Padding region is not blank, setting to blank")
                    value[: self.padding, :, class_index] = 0
                    value[-self.padding :, :, class_index] = 0
                    value[:, : self.padding, class_index] = 0
                    value[:, -self.padding :, class_index] = 0
        except IndexError as e:
            LOGGER.error(f"[{self.filename}] : Error mask is missing layers.", exc_info=e)

        # Update background class in case the mask has been edited
        value = update_background_class(value)
        self._mask: npt.NDArray[np.bool_] = value

    @property
    def padding(self) -> int:
        """
        Getter for the ``padding`` attribute.

        Returns
        -------
        int
            The padding amount.
        """
        return self._padding

    @padding.setter
    def padding(self, value: int) -> None:
        """
        Setter for the ``padding`` attribute.

        Parameters
        ----------
        value : int
            Padding amount.

        Raises
        ------
        ValueError
            If the padding is not an integer or is less than 1.
        """
        if not isinstance(value, int):
            raise ValueError(f"Padding must be an integer, but is {value}")
        if value < 1:
            raise ValueError(f"Padding must be >= 1, but is {value}")
        self._padding = value

    @property
    def bbox(self) -> tuple[int, int, int, int]:
        """
        Getter for the bounding box attribute.

        Returns
        -------
        tuple
            Bounding box of the crop.

        Raises
        ------
        ValueError
            If the bounding box is not square.
        """
        return self._bbox

    @bbox.setter
    def bbox(self, value: tuple[int, int, int, int]) -> None:
        """
        Setter for the bounding box attribute.

        Parameters
        ----------
        value : tuple[int, int, int, int]
            Bounding box of the crop.
        """
        if len(value) != 4:
            raise ValueError(f"Bounding box must have 4 elements, but has {len(value)}")
        if value[2] - value[0] != value[3] - value[1]:
            raise ValueError(
                f"Bounding box is not square: {value}, size: {value[2] - value[0]} x {value[3] - value[1]}"
            )
        self._bbox = value

    @property
    def pixel_to_nm_scaling(self) -> float:
        """
        Getter for the pixel to nanometre scaling factor attribute.

        Returns
        -------
        float
            Pixel to nanometre scaling factor.
        """
        return self._pixel_to_nm_scaling

    @pixel_to_nm_scaling.setter
    def pixel_to_nm_scaling(self, value: float) -> None:
        """
        Setter for the pixel to nanometre scaling factor attribute.

        Parameters
        ----------
        value : float
            Pixel to nanometre scaling factor.
        """
        self._pixel_to_nm_scaling = value

    @property
    def thresholds(self) -> list[float]:
        """
        Getter for the ``thresholds`` attribute.

        Returns
        -------
        list[float]
            Returns the value of ``thresholds``.
        """
        return self._thresholds

    @thresholds.setter
    def thresholds(self, value: list[float]) -> None:
        """
        Setter for the ``thresholds`` attribute.

        Parameters
        ----------
        value : list[float]
            Value to set for ``thresholds``.
        """
        self._thresholds = value

    @property
    def filename(self) -> str:
        """
        Getter for the ``filename`` attribute.

        Returns
        -------
        str
            The image ``filename`` attribute.
        """
        return self._filename

    @filename.setter
    def filename(self, value: str) -> None:
        """
        Setter for the ``filename`` attribute.

        Parameters
        ----------
        value : str
            Image ``filename`` attribute.
        """
        self._filename = value

    @property
    def skeleton(self) -> npt.NDArray:
        """
        Getter for the ``skeleton`` attribute.

        Returns
        -------
        npt.NDArray
            Returns the value of ``skeleton``.
        """
        return self._skeleton

    @skeleton.setter
    def skeleton(self, value: npt.NDArray) -> None:
        """
        Setter for the ``skeleton`` attribute.

        Parameters
        ----------
        value : npt.NDArray
            Value to set for ``skeleton``.
        """
        self._skeleton = value

    @property
    def height_profiles(self) -> npt.NDArray:
        """
        Getter for the ``height_profile`` attribute.

        Returns
        -------
        str
            The image height_profile.
        """
        return self._height_profiles

    @height_profiles.setter
    def height_profiles(self, value: npt.NDArray) -> None:
        """
        Setter for the ``height_profile`` attribute.

        Parameters
        ----------
        value : str
            Image ``height_profile``.
        """
        self._height_profiles = value

    @property
    def stats(self) -> dict[str, Any]:
        """
        Getter for the stats.

        Returns
        -------
        str
            Dictionary of image statistics.
        """
        return self._stats

    @stats.setter
    def stats(self, value: dict[str, Any]) -> None:
        """
        Setter for the stats.

        Parameters
        ----------
        value : dict[str, Any]
            Image stats.
        """
        self._stats = value

    @property
    def disordered_trace(self) -> DisorderedTrace:
        """
        Getter for the ``disordered_trace`` attribute.

        Returns
        -------
        dict[str: Any]
            Returns the value of ``disordered_trace``.
        """
        return self._disordered_trace

    @disordered_trace.setter
    def disordered_trace(self, value: DisorderedTrace) -> None:
        """
        Setter for the ``disordered_trace`` attribute.

        Parameters
        ----------
        value : dict[str: Any]
            Value to set for ``disordered_trace``.
        """
        self._disordered_trace = value

    @property
    def nodes(self) -> dict[int, Node]:
        """
        Getter for the ``nodes`` attribute.

        Returns
        -------
        dict[int, Nodes]
            Returns ``nodes``, a dictionary of Nodes.
        """
        return self._nodes

    @nodes.setter
    def nodes(self, value: Node) -> None:
        """
        Setter for the ``nodes`` attribute.

        Parameters
        ----------
        value : Nodes
            Value to set for ``nodes``.
        """
        self._nodes = value

    @property
    def ordered_trace(self) -> OrderedTrace:
        """
        Getter for the ``ordered_trace`` attribute.

        Returns
        -------
        dict[str: Any]
            Returns the value of ``ordered_trace``.
        """
        return self._ordered_trace

    @ordered_trace.setter
    def ordered_trace(self, value: OrderedTrace) -> None:
        """
        Setter for the ``ordered_trace`` attribute.

        Parameters
        ----------
        value : dict[str: Any]
            Value to set for ``ordered_trace``.
        """
        self._ordered_trace = value

    @property
    def threshold_method(self) -> list[float]:
        """
        The ``threshold_method`` used to find the grain.

        Returns
        -------
        list[float]
            Returns the value of ``threshold_method``.
        """
        return self._threshold_method

    @threshold_method.setter
    def threshold_method(self, value: list[float]) -> None:
        """
        Setter for the ``threshold_method`` attribute.

        Parameters
        ----------
        value : list[float]
            Value to set for ``threshold_method``.
        """
        self._threshold_method = value

    def debug_locate_difference(self, other: object) -> None:  # noqa: C901 # pylint: disable=too-many-branches
        """
        Debug function to find the culprit when two GrainCrop objects are not equal.

        Parameters
        ----------
        other : object
            Object to compare to.

        Raises
        ------
        ValueError
            If the objects are not equal
        """
        if not isinstance(other, GrainCrop):
            raise ValueError(f"Cannot compare GrainCrop with {type(other)}")
        if not np.array_equal(self.image, other.image):
            raise ValueError(f"Image is different\n self.image  : {self.image}\n other.image : {other.image}")
        if not np.array_equal(self.mask, other.mask):
            raise ValueError(f"Mask is different\n self.mask  : {self.mask}\n other.mask : {other.mask}")
        if self.padding != other.padding:
            raise ValueError(f"Padding is different\n self.padding  : {self.padding}\n other.padding : {other.padding}")
        if self.bbox != other.bbox:
            raise ValueError(f"Bounding box is different\n self.bbox  : {self.bbox}\n other.bbox : {other.bbox}")
        if self.pixel_to_nm_scaling != other.pixel_to_nm_scaling:
            raise ValueError(
                "Pixel to nm scaling is different\n"
                f" self.pixel_to_nm_scaling  : {self.pixel_to_nm_scaling}\n"
                f" other.pixel_to_nm_scaling : {other.pixel_to_nm_scaling}"
            )
        if self.thresholds != other.thresholds:
            raise ValueError(
                f"Thresholds differ\n self.thresholds  : {self.thresholds}\n other.thresholds : {other.thresholds}"
            )
        if self.filename != other.filename:
            raise ValueError(
                f"Filename is different\n self.filename  : {self.filename}\n other.filename : {other.filename}"
            )
        if self.height_profiles != other.height_profiles:
            raise ValueError(
                "Height profiles are different\n"
                f" self.height_profiles  : {self.height_profiles}\n"
                f" other.height_profiles : {other.height_profiles}"
            )
        if self.skeleton != other.skeleton:
            raise ValueError(
                f"Skeleton is different\n self.skeleton  : {self.skeleton}\n other.skeleton : {other.skeleton}"
            )
        if self.disordered_trace != other.disordered_trace:
            raise ValueError(
                "Disordered traces are different\n"
                f" self.disordered_trace  : {self.disordered_trace}\n"
                f" other.disordered_trace : {other.disordered_trace}"
            )
        if self.nodes != other.nodes:
            raise ValueError(f"Nodes are different\n self.nodes  : {self.nodes}\n other.nodes : {other.nodes}")
        if self.threshold_method != other.threshold_method:
            raise ValueError(
                "Threshold Method is different\n"
                f" self.threshold_method  : {self.threshold_method}\n"
                f" other.threshold_method : {other.threshold_method}"
            )
        LOGGER.info("Cannot find difference between graincrops")


def validate_full_mask_tensor_shape(array: npt.NDArray[np.bool_]) -> npt.NDArray[np.bool_]:
    """
    Validate the shape of the full mask tensor.

    Parameters
    ----------
    array : npt.NDArray
        Numpy array to validate.

    Returns
    -------
    npt.NDArray
        Numpy array if valid.
    """
    if len(array.shape) != 3 or array.shape[2] < 2:
        raise ValueError(f"Full mask tensor must be WxHxC with C >= 2 but has shape {array.shape}")
    return array


@dataclass(
    repr=True, eq=True, config=ConfigDict(arbitrary_types_allowed=True, validate_assignment=True), validate_on_init=True
)
class DisorderedTrace:
    """
    Dataclass for storing the disordered tracing data.

    Attributes
    ----------
    images : dict[str: npt.NDArray]
        Dictionary of images generated during disordered tracing, should include ''pruned_skeleton''.
    grain_endpoints : int
        Number of Grain endpoints.
    grain_junctions : int
        Number of Grain junctions.
    total_branch_length : float
        Total branch length in nanometres.
    grain_width_mean : float
        Mean grain width in nanometres.
    stats : dict[int, Any]
        Dictionary of stats for each branch of a grain.
    """

    images: dict[str, npt.NDArray] | None = None
    grain_endpoints: int | None = None
    grain_junctions: int | None = None
    total_branch_length: float | None = None
    grain_width_mean: float | None = None
    stats: dict[int, Any] | None = None

    def __str__(self) -> str:
        """
        Representation function for useful statistics for the class.

        Returns
        -------
        str
            Set of formatted statistics on matched branches.
        """
        image_gens = ", ".join(self.images.keys())
        return (
            f"generated images : {image_gens}\n"
            f"grain endpoints : {self.grain_endpoints}\n"
            f"grain junctions : {self.grain_junctions}\n"
            f"total branch length (nm) : {self.total_branch_length}\n"
            f"mean grain width (nm) : {self.grain_width_mean}\n"
            f"number of branches : {len(self.stats)}"
        )


@dataclass(
    repr=True, eq=True, config=ConfigDict(arbitrary_types_allowed=True, validate_assignment=True), validate_on_init=True
)
class TopoStats:
    """
    Class for storing TopoStats objects.

    Attributes
    ----------
    grain_crops : dict[int, GrainCrop] | None
        Dictionary of ``GrainCrop`` objects.
    filename : str | None
        Filename.
    image_name : str | None
        Filename without its extension.
    pixel_to_nm_scaling : str | None
        Pixel to nanometre scaling.
    img_path : str | None
        Original path to image.
    image : npt.NDArray | None
        Flattened image (post ``Filter()``).
    image_original : npt.NDArray | None
        Original image.
    image_statistics : pd.DataFrame | None
        Pandas dataframe of
    full_mask_tensor : npt.NDArray
        Tensor mask for the full image.
    topostats_version : str | None
        TopoStats version.
    config : dict[str, Any] | None
        Configuration used when processing the grain.
    full_image_plots : dict[str, npt.NDArray[np.float64]]
        Dictionary of numpy arrays where disordered tracing, nodestats and ordered tracing skeletons are mapped back to
        the original image.
    basename : str
        Basename of image locations.
    """

    grain_crops: dict[int, GrainCrop] | None = None
    filename: str | None = None
    image_name: str | None = None
    pixel_to_nm_scaling: float | None = None
    img_path: Path | str | None = None
    image: npt.NDArray | None = None
    image_original: npt.NDArray | None = None
    image_statistics: dict[str, int | float | str] | None = None
    full_mask_tensor: npt.NDArray | None = None
    topostats_version: str | None = None
    config: dict[str, Any] | None = None
    full_image_plots: dict[str, npt.NDArray[np.float64]] | None = None
    basename: str | None = None

    def __str__(self) -> str:
        """
        Representation function for useful statistics for the class.

        Returns
        -------
        str
            Set of formatted and user-friendly statistics.
        """
        return (
            f"number of grain crops : {len(self.grain_crops)}\n"
            f"filename : {self.filename}\n"
            f"pixel to nm scaling : {self.pixel_to_nm_scaling}\n"
            f"image shape (px) : {self.image.shape}\n"
            f"image path : {self.img_path}\n"
            f"TopoStats version : {self.topostats_version}\n"
            f"Basename : {self.basename}"
        )

    def __eq__(self, other: object) -> bool:
        """
        Check if two ``TopoStats`` objects are equal.

        Parameters
        ----------
        other : object
            Other ``TopoStats`` object to compare to.

        Returns
        -------
        bool
            ``True`` if the objects are equal, ``False`` otherwise.
        """
        if not isinstance(other, TopoStats):
            return False
        return (
            self.grain_crops == other.grain_crops
            and self.filename == other.filename
            and self.pixel_to_nm_scaling == other.pixel_to_nm_scaling
            and self.img_path == other.img_path
            and self.topostats_version == other.topostats_version
            and self.config == other.config
            and np.array_equal(self.image, other.image)
            and np.array_equal(self.image_original, other.image_original)
            and np.array_equal(self.full_mask_tensor, other.full_mask_tensor)
        )

    def calculate_image_statistics(self) -> dict[str, int | float]:
        """
        Calculate the image statistics via ``statistics.image_statistics()``.

        Returns
        -------
        dict[str, int | float]
            Dictionary of image size and area in both metres and pixels, the number of grains, density and the
            root mean square of the image roughness.
        """
        self.image_statistics = image_statistics(
            image=self.image,
            filename=self.filename,
            pixel_to_nm_scaling=self.pixel_to_nm_scaling,
            n_grains=len(self.grain_crops) if self.grain_crops is not None and len(self.grain_crops) > 0 else 0,
        )
        return self.image_statistics


@dataclass(
    repr=True, eq=True, config=ConfigDict(arbitrary_types_allowed=True, validate_assignment=True), validate_on_init=True
)
class MatchedBranch:
    """
    Class for storing matched branches data and attributes.

    Attributes
    ----------
    ordered_coords : npt.NDArray[np.int32]
        Numpy array of ordered coordinates.
    heights : npt.NDArray[np.number]
        Numpy array of heights.
    distances : npt.NDArray[np.number]
        Numpy array of distances.
    fwhm : float
        Full-width half maximum.
    fwhm_half_maxs : list[float]
        Half-maximums from a matched branch.
    fwhm_peaks : list[int | float]
        Peaks from a matched branch.
    angles : float
        Angle between branches.
    branch_statistics : dict[str, float | int | list[Any] | str]
        Dictionary of branch statistics, ``fwhm``, ``fwhm_half_maxs`` and ``fwhm_peaks``.
    """

    ordered_coords: npt.NDArray[np.int32] | None = None
    heights: npt.NDArray[np.number] | None = None
    distances: npt.NDArray[np.number] | None = None
    fwhm: float | None = None
    fwhm_half_maxs: list[float] | None = None
    fwhm_peaks: list[float] | None = None
    angles: float | list[float] | None = None
    branch_statistics: dict[str, float | int | list[Any] | str] | None = None

    def __str__(self) -> str:
        """
        Representation function for useful statistics for the class.

        Returns
        -------
        str
            Set of formatted and user-friendly statistics.
        """
        return (
            f"number of coords : {self.ordered_coords.shape[0]}\n"
            f"full width half maximum : {self.fwhm}\n"
            f"full width half maximum maximums : {self.fwhm_half_maxs}\n"
            f"full width half maximum peaks : {self.fwhm_peaks}\n"
            f"angles : {self.angles}"
        )

    def collate_branch_statistics(self, image: str, basename: str) -> dict[str, float | int | list[Any] | str]:
        """
        Collate matched branch statistics to dictionary.

        Parameters
        ----------
        image : str
            Image being processed, added to dictionary for subsequent aggregation, typically ``TopoStats.filename``.
        basename : str
            Base image path, added to dictionary for subsequent aggregation, typically ``TopoStats.img_path``.

        Returns
        -------
        dict[str, float | int | list[Any] | str]
            Dictionary of ``fwhm``, ``fwhm_half_maxs`` and ``fwhm_peaks`` along with ``image`` and ``basename``.
        """
        self.branch_statistics = {
            "image": image,
            "fwhm": self.fwhm,
            "fwhm_half_maxs": self.fwhm_half_maxs,
            "fwhm_peaks": self.fwhm_peaks,
            "basename": str(basename),
        }
        return self.branch_statistics


@dataclass(
    repr=True, eq=True, config=ConfigDict(arbitrary_types_allowed=True, validate_assignment=True), validate_on_init=True
)
class UnMatchedBranch:
    """
    Class for storing matched branches data and attributes.

    Attributes
    ----------
    angles : float
        Angle between branches.
    """

    angles: float | list[float] | None = None

    def __str__(self) -> str:
        """
        Representation function for useful statistics for the class.

        Returns
        -------
        str
            Set of formatted and user-friendly statistics.
        """
        return f"angles : {self.angles}"


@dataclass(
    repr=True, eq=True, config=ConfigDict(arbitrary_types_allowed=True, validate_assignment=True), validate_on_init=True
)
class Node:
    """
    Class for storing Node data and attributes.

    Attributes
    ----------
    error : bool
        Whether an error occurred calculating statistics for this node.
    pixel_to_nm_scaling : np.float64
        Pixel to nanometre scaling.
    branch_stats : dict[int, MatchedBranch]
        Dictionary of ``MatchedBranch`` objects.
    unmatched_branch_stats : dict[int, UnMatchedBranch]
        Dictionary of ``UnmatchedBranch`` objects.
    node_coords : npt.NDArray[np.int32]
        Numpy array of node coordinates.
    confidence : np.float64
        Normalised confidence of the crossing in the range of ``0-1``.
    reduced_node_area : npt.NDArray[np.int32]
        The molecule skeleton, with all branches removed that are not connected to the current node.
    node_area_skeleton : npt.NDArray[np.int32]
        Numpy array of skeleton.
    node_branch_mask : npt.NDArray[np.int32]
        Numpy array of branch mask.
    node_avg_mask : npt.NDArray[np.int32]
        Numpy array of averaged mask.
    writhe : str
        Writhe type for the node.
    """

    error: bool | None = None
    pixel_to_nm_scaling: float | None = None
    branch_stats: dict[int, MatchedBranch] | None = None
    unmatched_branch_stats: dict[int, UnMatchedBranch] | None = None
    node_coords: npt.NDArray[np.int32] | None = None
    confidence: float | None = None
    reduced_node_area: npt.NDArray[np.int32] | None = None
    node_area_skeleton: npt.NDArray[np.int32] | None = None
    node_branch_mask: npt.NDArray[np.int32] | None = None
    node_avg_mask: npt.NDArray[np.int32] | None = None
    writhe: str | None = None

    def __str__(self) -> str:
        """
        Representation function for useful statistics for the class.

        Returns
        -------
        str
            Set of formatted and user-friendly statistics.
        """
        return (
            f"error : {self.error}\n"
            f"pixel to nm scaling (nm) : {self.pixel_to_nm_scaling}\n"
            f"number of matched branches : {len(self.branch_stats)}\n"
            f"number of unmatched branches : {len(self.unmatched_branch_stats)}\n"
            f"number of coords : {self.node_coords.shape[0]}\n"
            f"confidence : {self.confidence}\n"
            f"reduced node area : {self.reduced_node_area}"
        )


@dataclass(
    repr=True, eq=True, config=ConfigDict(arbitrary_types_allowed=True, validate_assignment=True), validate_on_init=True
)
class OrderedTrace:
    """
    Class for Ordered Trace data and attributes.

    molecule_data : dict[int, Molecule]
        Dictionary of ``Molecule`` objects indexed by molecule number.
    tracing_stats : dict | None
        Tracing statistics.
    grain_molstats : Any | None
        Grain molecule statistics.
    molecules : int
        Number of molecules within the grain.
    writhe : str
        The writhe sign, can be either `+`, `-` or `0` for positive, negative or no writhe.
    pixel_to_nm_scaling: np.float64 | None
        Pixel to nm scaling.
    images: dict[str, npt.NDArray] | None
        Diagnostic images produced during processing.
    error: bool | None
        Errors encountered?
    molecule_statistics : dict[int, dict[str, bool | str | float | None]] | None
        Dictionary of molecule statistics, with one entry for each molecule.
    """

    molecule_data: dict[int, Molecule] | None = None
    tracing_stats: dict | None = None
    grain_molstats: Any | None = None
    molecules: int | None = None
    writhe: str | None = None
    pixel_to_nm_scaling: float | None = None
    images: dict[str, npt.NDArray] | None = None
    error: bool | None = None
    molecule_statistics: dict[int, dict[str, bool | str | float | None]] | None = None

    def __str__(self) -> str:
        """
        Representation function for useful statistics for the class.

        Returns
        -------
        str
            Set of formatted and user-friendly statistics.
        """
        writhe = {"+": "positive", "-": "negative", "0": "no writhe"}.get(self.writhe)
        return (
            f"number of molecules : {self.molecules}\n"
            f"number of images : {len(self.images)}\n"
            f"writhe : {writhe}\n"
            f"pixel to nm scaling : {self.pixel_to_nm_scaling}\n"
            f"error : {self.error}"
        )

    def collate_molecule_statistics(self) -> dict[int, dict[str, bool | int | str | None]]:
        """
        Collate molecule statistics for all molecules to dictionary.

        The resulting dictionary can be easily converted to Pandas Dataframes for writing to CSV.

        Returns
        -------
        dict[int, dict[str, bool | int | str | None]]
            Dictionary, indexed by molecule where the value is the molecules statistics for the given molecule.
        """
        self.molecule_statistics = {
            molecule_number: molecule.collate_molecule_statistics()
            for molecule_number, molecule in self.molecule_data.items()
        }
        return self.molecule_statistics


@dataclass(
    repr=True, eq=True, config=ConfigDict(arbitrary_types_allowed=True, validate_assignment=True), validate_on_init=True
)
class Molecule:
    """
    Class for Molecules identified during ordered tracing.

    threshold : str
        Direction from threshold of molecule (above / below)
    molecule_number : int
        Index of the molecule (per grain)
    circular : str, bool, optional
        Whether the molecule is circular or linear.
    processing : str
        Which processing type was used, topostats or nodestats.
    topology : str, optional
        Topological classification of the molecule.
    topology_flip : Any, optional
        Unknown?
    ordered_coords : npt.NDArray, optional
        Ordered coordinates for the molecule.
    splined_coords : npt.NDArray, optional
        Smoothed (aka splined) coordinates for the molecule.
    contour_length : float
        Length of the molecule.
    end_to_end_distance : float
        Distance between ends of molecule. Will be ``0.0`` for circular molecules which don't have ends.
    heights : npt.NDArray
        Height along molecule.
    distances : npt.NDArray
        Distance between points on the molecule.
    curvature_stats : npt.NDArray, optional
        Angle changes along molecule. NB - These are always positive due to use of ``np.abs()`` during calculation.
    bbox : tuple[int, int, int, int], optional
        Bounding box.
    """

    threshold: str | None = None
    molecule_number: int | None = None
    circular: str | bool | None = None
    processing: str | None = None
    topology: str | None = None
    topology_flip: Any | None = None
    ordered_coords: npt.NDArray | None = None
    splined_coords: npt.NDArray | None = None
    contour_length: float | None = None
    end_to_end_distance: float | None = None
    heights: npt.NDArray | None = None
    distances: npt.NDArray | None = None
    curvature_stats: npt.NDArray | None = None
    bbox: tuple[int, int, int, int] | None = None
    molecule_statistics: dict[str, bool | str | float | None] | None = None

    def __str__(self) -> str:
        """
        Representation function for useful statistics for the class.

        Returns
        -------
        str
            Set of formatted and user-friendly statistics.
        """
        return (
            f"circular : {self.circular}\n"
            f"topology : {self.topology}\n"
            f"topology flip : {self.topology_flip}\n"
            f"number of ordered coords : {self.ordered_coords.shape[0]}\n"
            f"number of spline coords : {self.splined_coords.shape[0]}\n"
            f"contour length : {self.contour_length}\n"
            f"end to end distance : {self.end_to_end_distance}\n"
            f"bounding box coords : {self.bbox}"
        )

    def collate_molecule_statistics(self) -> dict[str, bool | str | float | None]:
        """
        Collate the molecule statsistics to a dictionary.

        Returns
        -------
        dict[str, bool | str | float | None]
            Dataframe of the classes attributes and their data.
        """
        self.molecule_statistics = {
            "circular": self.circular,
            "topology": self.topology,
            "topology_flip": self.topology_flip,
            "contour_length": self.contour_length,
            "end_to_end_distance": self.end_to_end_distance,
        }
        return self.molecule_statistics


def convert_to_dict(to_convert: Any) -> dict[str, Any]:
    """
    Convert a class to a dictionary representation of itself.

    Classes that are part of the ``topostats`` package (``TopoStats``, ``GrainCrop``, ``DisorderedTrace``, ``Node``,
    ``OrderedTrace``, ``MatchedBranch``, ``UnMatchedBranch`` and ``Molecule``) can be converted to dictionaries along
    with any other class that has the ``__dict__`` attribute.

    The keys are the attribute names (with the ``_`` prefix removed) and the values are the stored values of the
    attribute. Values of type ``str``, ``int``, ``float``, ``bool``, ``tuple``, ``np.ndarray`` and ``pathlib.Path`` are
    returned as is. If a value is itself a dictionary (e.g. ``TopoStats.grain_crops`` ``Node.branch_stats`` or
    ``OrderedTrace.molecule_data`` which are all dictionaries of the objects) then the object is recursed.

    If lists or dictionaries are provided a ``TypeError`` is raised and any object that doesn't have the ``__dict__``
    attribute then an ``AttributeError`` is raised.

    Parameters
    ----------
    to_convert : dict[str, Any]
        An object to be converted to a dictionary / dictionary item.

    Returns
    -------
    Any
        Input parameter as a dictionary / dictionary item.
    """
    if isinstance(to_convert, (int | float | str | Path | bool | np.ndarray | tuple | None)):
        return to_convert
    if isinstance(to_convert, list):
        return [convert_to_dict(item) for item in to_convert]
    if isinstance(to_convert, dict):
        return {key: convert_to_dict(value) for key, value in to_convert.items()}
    if hasattr(to_convert, "__dict__"):
        return {re.sub(r"^_", "", key): convert_to_dict(value) for key, value in to_convert.__dict__.items()}

    return to_convert



"""Topostats."""

import os
from importlib.metadata import version

import snoop
from matplotlib import colormaps
from packaging.version import Version

from .logs.logs import setup_logger
from .theme import Colormap

# Disable TensorFlow warnings
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"

LOGGER = setup_logger()

__version__ = version("topostats")
__release__ = ".".join(__version__.split(".")[:-2])

TOPOSTATS_VERSION = Version(__version__)
if TOPOSTATS_VERSION.is_prerelease and TOPOSTATS_VERSION.is_devrelease:
    TOPOSTATS_BASE_VERSION = str(TOPOSTATS_VERSION.base_version)
    TOPOSTATS_COMMIT = str(TOPOSTATS_VERSION).split("+g")[1]
else:
    TOPOSTATS_BASE_VERSION = str(TOPOSTATS_VERSION)
    TOPOSTATS_COMMIT = ""
CONFIG_DOCUMENTATION_REFERENCE = """# For more information on configuration and how to use it:
# https://afm-spm.github.io/TopoStats/main/configuration.html\n"""

colormaps.register(cmap=Colormap("nanoscope").get_cmap())
colormaps.register(cmap=Colormap("gwyddion").get_cmap())

# Disable snoop
snoop.install(enabled=False)


def log_topostats_version() -> None:
    """Log the TopoStats version, commit and date to system logger."""
    LOGGER.info(f"TopoStats version : {TOPOSTATS_VERSION}")
    LOGGER.info(f"Commit            : {TOPOSTATS_COMMIT}")



"""
Entry point for all TopoStats programs.

Parses command-line arguments and passes input on to the relevant functions / modules.
"""

import argparse as arg
import sys
from pathlib import Path

from topostats import __version__, log_topostats_version, run_modules
from topostats.config import update_module, write_config_with_comments
from topostats.plotting import run_toposum

# pylint: disable=too-many-lines
# pylint: disable=too-many-statements


def create_parser() -> arg.ArgumentParser:
    """
    Create a parser for reading options.

    Creates a parser, with multiple sub-parsers for reading options to run 'topostats'.

    Returns
    -------
    arg.ArgumentParser
        Argument parser.
    """
    parser = arg.ArgumentParser(
        description="Run various programs relating to AFM data. Add the name of the program you wish to run."
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"Installed version of TopoStats: {__version__}",
        help="Report the current version of TopoStats that is installed",
    )
    parser.add_argument(
        "-c",
        "--config-file",
        dest="config_file",
        type=Path,
        required=False,
        help="Path to a YAML configuration file.",
    )
    parser.add_argument(
        "-s",
        "--summary-config",
        dest="summary_config",
        required=False,
        help="Path to a YAML configuration file for summary plots and statistics.",
    )
    parser.add_argument(
        "--matplotlibrc",
        dest="matplotlibrc",
        type=Path,
        required=False,
        help="Path to a matplotlibrc file.",
    )
    parser.add_argument(
        "-b",
        "--base-dir",
        dest="base_dir",
        type=Path,
        required=False,
        help="Base directory to scan for images.",
    )
    parser.add_argument(
        "-o",
        "--output-dir",
        dest="output_dir",
        type=Path,
        required=False,
        help="Output directory to write results to.",
    )
    parser.add_argument(
        "-l",
        "--log-level",
        dest="log_level",
        type=str,
        required=False,
        help="Logging level to use, default is 'info' for verbose output use 'debug'.",
    )
    parser.add_argument(
        "-j",
        "--cores",
        dest="cores",
        type=int,
        required=False,
        help="Number of CPU cores to use when processing.",
    )
    parser.add_argument(
        "-f",
        "--file-ext",
        dest="file_ext",
        type=str,
        required=False,
        help="File extension to scan for.",
    )
    parser.add_argument(
        "--output-stats",
        dest="output_stats",
        type=str,
        required=False,
        help="'basic' (image and grain) or 'full' (image, grain, branch and molecule) statistics written to CSV.",
    )
    parser.add_argument(
        "--channel",
        dest="channel",
        type=str,
        required=False,
        help="Channel to extract.",
    )
    parser.add_argument(
        "--extract", dest="extract", type=str, required=False, help="Array to extract when loading '.topostats' files."
    )
    parser.add_argument(
        "--image-set",
        dest="image_set",
        type=list,
        nargs="+",
        required=False,
        help="Image set to generate, default is 'core' other options are 'all', "
        "'filters', 'grains', 'grain_crops', 'disordered_tracing', "
        "'nodestats', 'ordered_tracing', 'splining'.",
    )

    subparsers = parser.add_subparsers(title="program", description="Available programs, listed below:", dest="module")

    # Create a sub-parsers for different stages of processing and tasks
    process_parser = subparsers.add_parser(
        "process",
        description="Process AFM images. Additional arguments over-ride defaults or those in the configuration file.",
        help="Process AFM images. Additional arguments over-ride defaults or those in the configuration file.",
    )
    # Filter options
    process_parser.add_argument(
        "--filter-row-alignment-quantile",
        dest="filter_row_alignment_quantile",
        type=float,
        required=False,
        help="Lower values may improve flattening of larger features.",
    )
    process_parser.add_argument(
        "--filter-threshold-method",
        dest="filter_threshold_method",
        type=str,
        required=False,
        help="Method for thresholding Filtering. Options are otsu, std_dev, absolute.",
    )
    process_parser.add_argument(
        "--filter-otsu-threshold-multiplier",
        dest="filter_otsu_threshold_multiplier",
        type=float,
        required=False,
        help="Factor for scaling the Otsu threshold during Filtering.",
    )
    process_parser.add_argument(
        "--filter-threshold-std-dev-below",
        dest="filter_threshold_std_dev_below",
        type=float,
        required=False,
        help="Threshold for data below the image background for std dev method during Filtering.",
    )
    process_parser.add_argument(
        "--filter-threshold-std-dev-above",
        dest="filter_threshold_std_dev_above",
        type=float,
        required=False,
        help="Threshold for data above the image background for std dev method during Filtering.",
    )
    process_parser.add_argument(
        "--filter-threshold-absolute-below",
        dest="filter_threshold_absolute_below",
        type=float,
        required=False,
        help="Threshold for data below the image background dor absolute method during Filtering",
    )
    process_parser.add_argument(
        "--filter-threshold-absolute-above",
        dest="filter_threshold_absolute_above",
        type=float,
        required=False,
        help="Threshold for data above the image background dor absolute method during Filtering",
    )
    process_parser.add_argument(
        "--filter-gaussian-size",
        dest="filter_gaussian_size",
        type=float,
        required=False,
        help="Gaussian blur intensity in pixels.",
    )
    process_parser.add_argument(
        "--filter-gaussian-mode",
        dest="filter_gaussian_mode",
        type=str,
        required=False,
        help="Gaussian blur method. Options are 'nearest' (default), 'reflect', 'constant', 'mirror' or 'wrap'.",
    )
    process_parser.add_argument(
        "--filter-remove-scars", dest="filter_scars_run", type=bool, required=False, help="Whether to remove scars."
    )
    process_parser.add_argument(
        "--filter-scars-removal-iterations",
        dest="filter_scars_removal_iterations",
        type=int,
        required=False,
        help="Number of times to run scar removal",
    )
    process_parser.add_argument(
        "--filter-scars-threshold-low",
        dest="filter_scars_threshold_low",
        type=float,
        required=False,
        help="Lower values make scar removal more sensitive",
    )
    process_parser.add_argument(
        "--filter-scars-threshold-high",
        dest="filter_scars_threshold_high",
        type=float,
        required=False,
        help="Lower values make scar removal more sensitive",
    )
    process_parser.add_argument(
        "--filter-scars-max-scar-width",
        dest="filter_scars_max_scar_width",
        type=int,
        required=False,
        help="Maximum thickness of scars in pixels",
    )
    process_parser.add_argument(
        "--filter-scars-max-scar-length",
        dest="filter_scars_max_scar_length",
        type=int,
        required=False,
        help="Maximum length of scars in pixels",
    )

    # Grains
    process_parser.add_argument(
        "--grains-grain-crop-padding",
        dest="grains_grain_crop_padding",
        type=int,
        required=False,
        help="Padding to be applied while cropping grains, must be >= 1.",
    )
    process_parser.add_argument(
        "--grains-threshold-method",
        dest="grains_threshold_method",
        type=str,
        required=False,
        help="Method for thresholding Grain finding. Options are otsu, std_dev, absolute.",
    )
    process_parser.add_argument(
        "--grains-otsu-threshold-multiplier",
        dest="grains_otsu_threshold_multiplier",
        type=float,
        required=False,
        help="Factor for scaling the Otsu threshold during Grain finding.",
    )
    process_parser.add_argument(
        "--grains-threshold-std-dev-below",
        dest="grains_threshold_std_dev_below",
        type=float,
        required=False,
        help="Threshold for data below the image background for std dev method during Grain finding.",
    )
    process_parser.add_argument(
        "--grains-threshold-std-dev-above",
        dest="grains_threshold_std_dev_above",
        type=float,
        required=False,
        help="Threshold for data above the image background for std dev method during Grain finding.",
    )
    process_parser.add_argument(
        "--grains-threshold-absolute-below",
        dest="grains_threshold_absolute_below",
        type=float,
        required=False,
        help="Threshold for data below the image background dor absolute method during Grain finding",
    )
    process_parser.add_argument(
        "--grains-threshold-absolute-above",
        dest="grains_threshold_absolute_above",
        type=float,
        required=False,
        help="Threshold for data above the image background dor absolute method during Grain finding",
    )
    process_parser.add_argument(
        "--grains-direction",
        dest="grains_direction",
        type=str,
        required=False,
        help="Whether to look for grains 'above' pr 'below' thresholds of 'both'",
    )
    process_parser.add_argument(
        "--grains-area-thresholds-above",
        dest="grains_area_thresholds_above",
        type=float,
        required=False,
        nargs=2,
        help="Above surface (low, high) in nm^2, both low and high should be specified",
    )
    process_parser.add_argument(
        "--grains-area-thresholds-below",
        dest="grains_area_thresholds_below",
        type=float,
        required=False,
        nargs=2,
        help="Below surface (low, high) in nm^2, both low and high should be specified",
    )
    process_parser.add_argument(
        "--grains-remove-edge-intersecting-grains",
        dest="grains_remove_edge_intersecting_grains",
        type=bool,
        required=False,
        help="Whether or not to remove grains that touch the image border",
    )
    # Unet
    process_parser.add_argument(
        "--unet-model-path", dest="unet_model_path", type=Path, required=False, help="Path to a trained U-net model"
    )
    process_parser.add_argument(
        "--unet-grain-crop-padding",
        dest="unet_grain_crop_padding",
        type=int,
        required=False,
        help="Padding to apply to the grain crop bounding box",
    )
    process_parser.add_argument(
        "--unet-upper-norm-bound",
        dest="unet_upper_norm_bound",
        type=float,
        required=False,
        help="Upper bound for normalisation of input data. This should be slightly higher than the maximum desired"
        "/expected height of grains",
    )
    process_parser.add_argument(
        "--unet-lower-norm-bound",
        dest="unet_lower_norm_bound",
        type=float,
        required=False,
        help="Lower bound for normalisation of input data. This should be slightly lower than the minimum desired"
        "/expected height of the background",
    )
    process_parser.add_argument(
        "--unet-remove-disconnected-grains",
        dest="unet_remove_disconnected_grains",
        type=bool,
        required=False,
        help="Whether or not to remove grains that are disconnected from the original mask",
    )
    process_parser.add_argument(
        "--unet-confidence",
        dest="unet_confidence",
        type=float,
        required=False,
        help="Confidence threshold for the UNet model. Smaller is more generous, large is stricter.",
    )

    # Grainstats
    process_parser.add_argument(
        "--grainstats-edge-detection-method",
        dest="grainstats_edge_detection_method",
        type=str,
        required=False,
        help="Method of edge detection, do NOT change this unless you are sure of what it will do. Options 'canny' and"
        "'binary_erosion (default)",
    )
    process_parser.add_argument(
        "--grainstats-extract-height-profile",
        dest="grainstats_extract_height_profile",
        type=bool,
        required=False,
        help="Extract height profiles along maximum feret of molecules",
    )
    process_parser.add_argument(
        "--grainstats-class-names",
        dest="grainstats_class_names",
        type=list,
        nargs="+",
        required=False,
        help=(
            "Names of the objects found by the segmentation methods (objects "
            "are linked to the `class_number` output in the grainstats csv"
        ),
    )

    # Disordered Tracing
    process_parser.add_argument(
        "--disordered-class-index",
        dest="disordered_class_index",
        type=int,
        required=False,
        help="Index of the class to trace.",
    )
    process_parser.add_argument(
        "--disordered-min-skeleton-size",
        dest="disordered_min_skeleton_size",
        type=float,
        required=False,
        help="Minimum number of pixels in a skeleton for it to be retained",
    )
    process_parser.add_argument(
        "--disordered-mask-smoothing-params-gaussian-sigma",
        dest="disordered_mask_smoothing_params_gaussian_sigma",
        type=float,
        required=False,
        help="Gaussian smoothing parameter 'sigma' in pixels",
    )
    process_parser.add_argument(
        "--disordered-mask-smoothing-params-dilation-iterations",
        dest="disordered_mask_smoothing_params_dilation_iterations",
        type=int,
        required=False,
        help="Number of dilation iterations to use for grain smoothing",
    )
    process_parser.add_argument(
        "--disordered-mask-smoothing-params-holearea-min-max",
        dest="disordered_mask_smoothing_params_holearea_min_max",
        type=float,
        required=False,
        nargs=2,
        help="Range (min, max) of a hole area in nm to refil in the smoothed masks",
    )
    process_parser.add_argument(
        "--disordered-skeletonisation-params-method",
        dest="disordered_skeletonisation_params_method",
        type=str,
        required=False,
        help="Skeletonisation method. Options : zhang | lee | thin | topostats",
    )
    process_parser.add_argument(
        "--disordered-skeletonisation-height-bias",
        dest="disordered_skeletonisation_height_bias",
        type=float,
        required=False,
        help="Percentage of lowest pixels to remove each skeletonisation iteration. 1.0 equates to Zhang method",
    )
    process_parser.add_argument(
        "--disordered-pruning-params-method",
        dest="disordered_pruning_params_method",
        type=str,
        required=False,
        help="Method to clean branches from the skeleton: Options : 'topostats'",
    )
    process_parser.add_argument(
        "--disordered-pruning-params-max-length",
        dest="disordered_pruning_params_max_length",
        type=float,
        required=False,
        help="Maximum length in nm to remove a branch containing an endpoint",
    )
    process_parser.add_argument(
        "--disordered-pruning-params-height-threshold",
        dest="disordered_pruning_params_height_threshold",
        type=float,
        required=False,
        help="The height to remove branches below",
    )
    process_parser.add_argument(
        "--disordered-pruning-params-method-values",
        dest="disordered_pruning_params_method_values",
        type=str,
        required=False,
        help="Method to obtain a branch's height for pruning. Options : 'min' | 'median', 'mid'",
    )
    process_parser.add_argument(
        "--disordered-pruning-params-method-outlier",
        dest="disordered_pruning_params_method_outlier",
        type=str,
        required=False,
        help="Method to prune branches based on height. Options : 'abs' | 'mean_abs' | 'iqr'",
    )
    process_parser.add_argument(
        "--disordered-pruning-params-only-height-prune-endpoints",
        dest="disordered_pruning_params_only_height_prune_endpoints",
        type=bool,
        required=False,
        help="Restrict height-based pruning to skeleton segments containing an endpoint.",
    )

    # Nodestats
    process_parser.add_argument(
        "--nodestats-node-joining-length",
        dest="nodestats_node_joining_length",
        type=float,
        required=False,
        help="The distance over which to join nearby crossing points",
    )
    process_parser.add_argument(
        "--nodestats-node-extend-dist",
        dest="nodestats_node_extend_dist",
        type=float,
        required=False,
        help="The distance over which to join nearby odd-branched nodes",
    )
    process_parser.add_argument(
        "--nodestats-branch-pairing-length",
        dest="nodestats_branch_pairing_length",
        type=float,
        required=False,
        help="The length from the crossing point to pair and trace, obtaining FWHM's",
    )
    process_parser.add_argument(
        "--nodestats-pair-odd-branches",
        dest="nodestats_pair_odd_branches",
        type=bool,
        required=False,
        help="Whether to try and pair odd-branched nodes",
    )

    # Ordered Tracing
    process_parser.add_argument(
        "--ordered-ordering-method",
        dest="ordered_ordering_method",
        type=str,
        required=False,
        help="Ordering method for ordering disordered traces. Option 'nodestats'",
    )

    # Splining
    process_parser.add_argument(
        "--splining-method",
        dest="splining_method",
        type=str,
        required=False,
        help="Method for splining. Options 'spline' or 'rolling_window",
    )
    process_parser.add_argument(
        "--rolling-window-size",
        dest="rolling_window_size",
        type=float,
        required=False,
        help="Size in nm of the rolling window",
    )
    process_parser.add_argument(
        "--splining-resampling",
        dest="splining_resampling",
        type=bool,
        required=False,
        help="Whether to resample the trace or not.",
    )
    process_parser.add_argument(
        "--splining-resample-regular-spatial-interval",
        dest="splining_resample_regular_spatial_interval",
        type=float,
        required=False,
        help="The spatial interval to resample the trace to in nm.",
    )
    process_parser.add_argument(
        "--spline-step-size",
        dest="spline_step_size",
        type=float,
        required=False,
        help="The sampling rate of the spline in metres",
    )
    process_parser.add_argument(
        "--spline-linear-smoothing",
        dest="spline_linear_smoothing",
        type=float,
        required=False,
        help="The amount of smoothing to apply to linear features",
    )
    process_parser.add_argument(
        "--spline-circular-smoothing",
        dest="spline_circular_smoothing",
        type=float,
        required=False,
        help="The amount of smoothing to apply to circular features",
    )
    process_parser.add_argument(
        "--spline-degree",
        dest="spline_degree",
        type=int,
        required=False,
        help="The polynomial degree of the spline",
    )

    # Plotting
    process_parser.add_argument(
        "--save-plots",
        dest="save_plots",
        type=bool,
        required=False,
        help="Whether to save plots.",
    )
    process_parser.add_argument(
        "--savefig-format",
        dest="savefig_format",
        type=str,
        required=False,
        help="Format for saving figures to, options are 'png', 'svg', or other valid Matplotlib supported formats.",
    )
    process_parser.add_argument(
        "--savefig-dpi",
        dest="savefig_dpi",
        type=int,
        required=False,
        help="Dots Per Inch for plots, should be integer for dots per inch.",
    )
    process_parser.add_argument(
        "--cmap",
        dest="cmap",
        type=str,
        required=False,
        help="Colormap to use, options include 'nanoscope', 'afmhot' or any valid Matplotlib colormap.",
    )
    process_parser.add_argument(
        "--grain-crop-plot-size-nm",
        dest="grain_crop_plot_size_nm",
        type=float,
        required=False,
        help="Size in nm of the square cropped grain images if using the grains image set. If -1, will "
        "use the grain's default bounding box size.",
    )
    process_parser.add_argument("-m", "--mask", dest="mask", type=bool, required=False, help="Mask the image.")
    process_parser.add_argument(
        "-w",
        "--warnings",
        dest="warnings",
        type=bool,
        required=False,
        help="Whether to ignore warnings.",
    )
    process_parser.add_argument(
        "--number-masks",
        dest="number_grains",
        type=bool,
        required=False,
        help="Add numbers to each grain mask in outputted mask images.",
    )
    # Run the relevant function with the arguments
    process_parser.set_defaults(func=run_modules.process)

    load_parser = subparsers.add_parser(
        "load",
        description="Load and save all images as .topostats files for subsequent processing.",
        help="Load and save all images as .topostats files for subsequent processing.",
    )
    # Run the relevant function with the arguments
    load_parser.set_defaults(func=run_modules.process)

    # Filter
    filter_parser = subparsers.add_parser(
        "filter",
        description="Load and filter images, saving as .topostats files for subsequent processing.",
        help="Load and filter images, saving as .topostats files for subsequent processing.",
    )
    filter_parser.add_argument(
        "--row-alignment-quantile",
        dest="row_alignment_quantile",
        type=float,
        required=False,
        help="Lower values may improve flattening of larger features.",
    )
    filter_parser.add_argument(
        "--threshold-method",
        dest="threshold_method",
        type=str,
        required=False,
        help="Method for thresholding Filtering. Options are otsu, std_dev, absolute.",
    )
    filter_parser.add_argument(
        "--otsu-threshold-multiplier",
        dest="otsu_threshold_multiplier",
        type=float,
        required=False,
        help="Factor for scaling the Otsu threshold during Filtering.",
    )
    filter_parser.add_argument(
        "--threshold-std-dev-below",
        dest="threshold_std_dev_below",
        type=float,
        required=False,
        help="Threshold for data below the image background for std dev method during Filtering.",
    )
    filter_parser.add_argument(
        "--threshold-std-dev-above",
        dest="threshold_std_dev_above",
        type=float,
        required=False,
        help="Threshold for data above the image background for std dev method during Filtering.",
    )
    filter_parser.add_argument(
        "--threshold-absolute-below",
        dest="threshold_absolute_below",
        type=float,
        required=False,
        help="Threshold for data below the image background dor absolute method during Filtering",
    )
    filter_parser.add_argument(
        "--threshold-absolute-above",
        dest="threshold_absolute_above",
        type=float,
        required=False,
        help="Threshold for data above the image background dor absolute method during Filtering",
    )
    filter_parser.add_argument(
        "--gaussian-size", dest="gaussian_size", type=float, required=False, help="Gaussian blur intensity in pixels."
    )
    filter_parser.add_argument(
        "--gaussian-mode",
        dest="gaussian_mode",
        type=str,
        required=False,
        help="Gaussian blur method. Options are 'nearest' (default), 'reflect', 'constant', 'mirror' or 'wrap'.",
    )
    filter_parser.add_argument(
        "--remove-scars", dest="scars_run", type=bool, required=False, help="Whether to remove scars."
    )
    filter_parser.add_argument(
        "--scars-removal-iterations",
        dest="scars_removal_iterations",
        type=int,
        required=False,
        help="Number of times to run scar removal",
    )
    filter_parser.add_argument(
        "--scars-threshold-low",
        dest="scars_threshold_low",
        type=float,
        required=False,
        help="Lower values make scar removal more sensitive",
    )
    filter_parser.add_argument(
        "--scars-threshold-high",
        dest="scars_threshold_high",
        type=float,
        required=False,
        help="Lower values make scar removal more sensitive",
    )
    filter_parser.add_argument(
        "--scars-max-scar-width",
        dest="scars_max_scar_width",
        type=int,
        required=False,
        help="Maximum thickness of scars in pixels",
    )
    filter_parser.add_argument(
        "--scars-max-scar-length",
        dest="scars_max_scar_length",
        type=int,
        required=False,
        help="Maximum length of scars in pixels",
    )
    # Run the relevant function with the arguments
    filter_parser.set_defaults(func=run_modules.filters)

    grains_parser = subparsers.add_parser(
        "grains",
        description="Load filtered images from '.topostats' files and detect grains.",
        help="Load filtered images from '.topostats' files and detect grains.",
    )
    grains_parser.add_argument(
        "--grain-crop-padding",
        dest="grain_crop_padding",
        type=int,
        required=False,
        help="Padding to be applied while cropping grains, must be at least 1.",
    )
    grains_parser.add_argument(
        "--threshold-method",
        dest="threshold_method",
        type=str,
        required=False,
        help="Method for thresholding Grain finding. Options are otsu, std_dev, absolute.",
    )
    grains_parser.add_argument(
        "--otsu-threshold-multiplier",
        dest="otsu_threshold_multiplier",
        type=float,
        required=False,
        help="Factor for scaling the Otsu threshold during Grain finding.",
    )
    grains_parser.add_argument(
        "--threshold-std-dev-below",
        dest="threshold_std_dev_below",
        type=float,
        required=False,
        help="Threshold for data below the image background for std dev method during Grain finding.",
    )
    grains_parser.add_argument(
        "--threshold-std-dev-above",
        dest="threshold_std_dev_above",
        type=float,
        required=False,
        help="Threshold for data above the image background for std dev method during Grain finding.",
    )
    grains_parser.add_argument(
        "--threshold-absolute-below",
        dest="threshold_absolute_below",
        type=float,
        required=False,
        help="Threshold for data below the image background dor absolute method during Grain finding",
    )
    grains_parser.add_argument(
        "--threshold-absolute-above",
        dest="threshold_absolute_above",
        type=float,
        required=False,
        help="Threshold for data above the image background dor absolute method during Grain finding",
    )
    grains_parser.add_argument(
        "--direction",
        dest="direction",
        type=str,
        required=False,
        help="Whether to look for grains 'above' pr 'below' thresholds of 'both'",
    )
    grains_parser.add_argument(
        "--area-thresholds-above",
        dest="area_thresholds_above",
        type=float,
        required=False,
        nargs=2,
        help="Above surface (low, high) in nm^2, both low and high should be specified",
    )
    grains_parser.add_argument(
        "--area-thresholds-below",
        dest="area_thresholds_below",
        type=float,
        required=False,
        nargs=2,
        help="Below surface (low, high) in nm^2, both low and high should be specified",
    )
    grains_parser.add_argument(
        "--remove-edge-intersecting-grains",
        dest="remove_edge_intersecting_grains",
        type=bool,
        required=False,
        help="Whether or not to remove grains that touch the image border",
    )
    # Unet
    grains_parser.add_argument(
        "--unet-model-path", dest="unet_model_path", type=Path, required=False, help="Path to a trained U-net model"
    )
    grains_parser.add_argument(
        "--unet-grain-crop-padding",
        dest="unet_grain_crop_padding",
        type=int,
        required=False,
        help="Padding to apply to the grain crop bounding box",
    )
    grains_parser.add_argument(
        "--unet-upper-norm-bound",
        dest="unet_upper_norm_bound",
        type=float,
        required=False,
        help="Upper bound for normalisation of input data. This should be slightly higher than the maximum desired"
        "/expected height of grains",
    )
    grains_parser.add_argument(
        "--unet-lower-norm-bound",
        dest="unet_lower_norm_bound",
        type=float,
        required=False,
        help="Lower bound for normalisation of input data. This should be slightly lower than the minimum desired"
        "/expected height of the background",
    )
    grains_parser.add_argument(
        "--unet-remove-disconnected-grains",
        dest="unet_remove_disconnected_grains",
        type=bool,
        required=False,
        help="Whether or not to remove grains that are disconnected from the original mask",
    )
    grains_parser.add_argument(
        "--unet-confidence",
        dest="unet_confidence",
        type=float,
        required=False,
        help="Confidence threshold for the UNet model. Smaller is more generous, large is stricter.",
    )
    # Run the relevant function with the arguments
    grains_parser.set_defaults(func=run_modules.grains)

    grainstats_parser = subparsers.add_parser(
        "grainstats",
        description="Load images with grains from '.topostats' files and calculate statistics.",
        help="Load images with grains from '.topostats' files and calculate statistics.",
    )
    grainstats_parser.add_argument(
        "--edge-detection-method",
        dest="edge_detection_method",
        type=str,
        required=False,
        help="Method of edge detection, do NOT change this unless you are sure of what it will do. Options 'canny' and"
        "'binary_erosion (default)",
    )
    grainstats_parser.add_argument(
        "--extract-height-profile",
        dest="extract_height_profile",
        type=bool,
        required=False,
        help="Extract height profiles along maximum feret of molecules",
    )
    grainstats_parser.add_argument(
        "--class-names",
        dest="class_names",
        type=list,
        nargs="+",
        required=False,
        help=(
            "Names of the objects found by the segmentation methods (objects "
            "are linked to the `class_number` output in the grainstats csv"
        ),
    )
    # Run the relevant function with the arguments
    grainstats_parser.set_defaults(func=run_modules.grainstats)

    # Disordered
    disordered_tracing_parser = subparsers.add_parser(
        "disordered-tracing",
        description="WIP DO NOT USE - Skeletonise and prune objects to disordered traces.",
        help="WIP DO NOT USE - Skeletonise and prune objects to disordered traces.",
    )
    disordered_tracing_parser.add_argument(
        "--class-index",
        dest="class_index",
        type=int,
        required=False,
        help="Index of the class to trace",
    )
    disordered_tracing_parser.add_argument(
        "--min-skeleton-size",
        dest="min_skeleton_size",
        type=float,
        required=False,
        help="Minimum number of pixels in a skeleton for it to be retained",
    )
    disordered_tracing_parser.add_argument(
        "--mask-smoothing-params-gaussian-sigma",
        dest="mask_smoothing_params_gaussian_sigma",
        type=float,
        required=False,
        help="Gaussian smoothing parameter 'sigma' in pixels",
    )
    disordered_tracing_parser.add_argument(
        "--mask-smoothing-params-dilation-iterations",
        dest="mask_smoothing_params_dilation_iterations",
        type=int,
        required=False,
        help="Number of dilation iterations to use for grain smoothing",
    )
    disordered_tracing_parser.add_argument(
        "--mask-smoothing-params-holearea-min-max",
        dest="mask_smoothing_params_holearea_min_max",
        type=float,
        required=False,
        nargs=2,
        help="Range (min, max) of a hole area in nm to refil in the smoothed masks",
    )
    disordered_tracing_parser.add_argument(
        "--skeletonisation-params-method",
        dest="skeletonisation_params_method",
        type=str,
        required=False,
        help="Skeletonisation method. Options : zhang | lee | thin | topostats",
    )
    disordered_tracing_parser.add_argument(
        "--skeletonisation-height-bias",
        dest="skeletonisation_height_bias",
        type=float,
        required=False,
        help="Percentage of lowest pixels to remove each skeletonisation iteration. 1.0 equates to Zhang method",
    )
    disordered_tracing_parser.add_argument(
        "--pruning-params-method",
        dest="pruning_params_method",
        type=str,
        required=False,
        help="Method to clean branches from the skeleton: Options : 'topostats'",
    )
    disordered_tracing_parser.add_argument(
        "--pruning-params-max-length",
        dest="pruning_params_max_length",
        type=float,
        required=False,
        help="Maximum length in nm to remove a branch containing an endpoint",
    )
    disordered_tracing_parser.add_argument(
        "--pruning-params-height-threshold",
        dest="pruning_params_height_threshold",
        type=float,
        required=False,
        help="The height to remove branches below",
    )
    disordered_tracing_parser.add_argument(
        "--pruning-params-method-values",
        dest="pruning_params_method_values",
        type=str,
        required=False,
        help="Method to obtain a branch's height for pruning. Options : 'min' | 'median' | 'mid'",
    )
    disordered_tracing_parser.add_argument(
        "--pruning-params-method-outlier",
        dest="pruning_params_method_outlier",
        type=str,
        required=False,
        help="Method to prune branches based on height. Options : 'abs' | 'mean_abs' | 'iqr'",
    )
    disordered_tracing_parser.add_argument(
        "--pruning-params-only-height-prune-endpoints",
        dest="pruning_params_only_height_prune_endpoints",
        type=bool,
        required=False,
        help="Restrict height-based pruning to skeleton segments containing an endpoint.",
    )
    # Run the relevant function with the arguments
    disordered_tracing_parser.set_defaults(func=run_modules.disordered_tracing)

    # Nodestats
    nodestats_parser = subparsers.add_parser(
        "nodestats",
        description="WIP DO NOT USE - Calculate node statistics and disentangle molecules.",
        help="WIP DO NOT USE - Calculate node statistics and disentangle molecules.",
    )
    nodestats_parser.add_argument(
        "--node-joining-length",
        dest="node_joining_length",
        type=float,
        required=False,
        help="The distance over which to join nearby crossing points",
    )
    nodestats_parser.add_argument(
        "--node-extend-dist",
        dest="node_extend_dist",
        type=float,
        required=False,
        help="The distance over which to join nearby odd-branched nodes",
    )
    nodestats_parser.add_argument(
        "--branch-pairing-length",
        dest="branch_pairing_length",
        type=float,
        required=False,
        help="The length from the crossing point to pair and trace, obtaining FWHM's",
    )
    nodestats_parser.add_argument(
        "--pair-odd-branches",
        dest="pair_odd_branches",
        type=bool,
        required=False,
        help="Whether to try and pair odd-branched nodes",
    )
    # Run the relevant function with the arguments
    nodestats_parser.set_defaults(func=run_modules.nodestats)

    # Ordered Tracing
    ordered_tracing_parser = subparsers.add_parser(
        "ordered-tracing",
        description="WIP DO NOT USE - Ordered traces of pruned skeletons.",
        help="WIP DO NOT USE - Ordered traces of pruned skeletons.",
    )
    ordered_tracing_parser.add_argument(
        "--ordering-method",
        dest="ordering_method",
        type=str,
        required=False,
        help="Ordering method for ordering disordered traces. Option 'nodestats'",
    )
    # Run the relevant function with the arguments
    ordered_tracing_parser.set_defaults(func=run_modules.ordered_tracing)

    # Splining
    splining_parser = subparsers.add_parser(
        "splining",
        description="WIP DO NOT USE - Splining of traced molecules to produce smooth curves.",
        help="WIP DO NOT USE - Splining of traced molecules to produce smooth curves.",
    )
    splining_parser.add_argument(
        "--method",
        dest="method",
        type=str,
        required=False,
        help="Method for splining. Options 'spline' or 'rolling_window",
    )
    splining_parser.add_argument(
        "--rolling-window-size",
        dest="rolling_window_size",
        type=float,
        required=False,
        help="Size in nm of the rolling window",
    )
    splining_parser.add_argument(
        "--resampling",
        dest="resampling",
        type=bool,
        required=False,
        help="Whether to resample the trace or not.",
    )
    splining_parser.add_argument(
        "--resample-regular-spatial-interval",
        dest="resample_regular_spatial_interval",
        type=float,
        required=False,
        help="The spatial interval to resample the trace to in nm.",
    )
    splining_parser.add_argument(
        "--spline-step-size",
        dest="spline_step_size",
        type=float,
        required=False,
        help="The sampling rate of the spline in metres",
    )
    splining_parser.add_argument(
        "--spline-linear-smoothing",
        dest="spline_linear_smoothing",
        type=float,
        required=False,
        help="The amount of smoothing to apply to linear features",
    )
    splining_parser.add_argument(
        "--spline-circular-smoothing",
        dest="spline_circular_smoothing",
        type=float,
        required=False,
        help="The amount of smoothing to apply to circular features",
    )
    splining_parser.add_argument(
        "--spline-degree",
        dest="spline_degree",
        type=int,
        required=False,
        help="The polynomial degree of the spline",
    )
    # Run the relevant function with the arguments
    splining_parser.set_defaults(func=run_modules.splining)

    summary_parser = subparsers.add_parser(
        "summary",
        description="Plotting and summary of TopoStats output statistics.",
        help="Plotting and summary of TopoStats output statistics.",
    )
    summary_parser.add_argument(
        "--input-csv",
        dest="input_csv",
        type=Path,
        required=False,
        help="Path to CSV file to plot.",
    )
    summary_parser.add_argument(
        "--config-file",
        dest="config_file",
        type=Path,
        required=False,
        help="Path to a YAML plotting dictionary that maps variable names to labels.",
    )
    summary_parser.add_argument(
        "--var-to-label",
        dest="var_to_label",
        type=Path,
        required=False,
        help="Path to a YAML plotting dictionary that maps variable names to labels.",
    )
    summary_parser.add_argument(
        "--create-config-file",
        dest="create_config_file",
        type=Path,
        required=False,
        help="Filename to write a sample YAML configuration file to (should end in '.yaml').",
    )
    summary_parser.add_argument(
        "--savefig-format",
        dest="savefig_format",
        type=str,
        required=False,
        help="Format for saving figures to, options are 'png', 'svg', or other valid Matplotlib supported formats.",
    )
    summary_parser.set_defaults(func=run_toposum)

    create_config_parser = subparsers.add_parser(
        "create-config",
        description="Create a configuration file using the defaults.",
        help="Create a configuration file using the defaults.",
    )
    create_config_parser.add_argument(
        "-f",
        "--filename",
        dest="filename",
        type=Path,
        required=False,
        help="Name of YAML file to save configuration to (default 'config.yaml').",
    )
    create_config_parser.add_argument(
        "-o",
        "--output-dir",
        dest="output_dir",
        type=Path,
        required=False,
        default="./",
        help="Path to where the YAML file should be saved (default './' the current directory).",
    )
    create_config_parser.add_argument(
        "-c",
        "--config",
        dest="config",
        type=str,
        default="default",
        help="Configuration to use, currently 'default', 'simple', 'mplstyle' and 'var_to_label' are supported.",
    )
    create_config_parser.set_defaults(func=write_config_with_comments)

    # Rename old Bruker files
    bruker_rename = subparsers.add_parser(
        "bruker-rename",
        description="Append the suffix '.spm' to old (numeric) Bruker file extensions.",
        help="Append the suffix '.spm' to old (numeric) Bruker file extensions.",
    )
    # Run the relevant function with the arguments
    bruker_rename.set_defaults(func=run_modules.bruker_rename)

    return parser


def entry_point(manually_provided_args=None, testing=False) -> None:
    """
    Entry point for all TopoStats programs.

    Main entry point for running 'topostats' which allows the different processing steps ('process', 'filter',
    'create_config' etc.) to be run.

    Parameters
    ----------
    manually_provided_args : None
        Manually provided arguments.
    testing : bool
        Whether testing is being carried out.

    Returns
    -------
    None
        Does not return anything.
    """
    # Log topostats version and the commit id
    log_topostats_version()

    # Parse command line options, load config (or default) and update with command line options
    parser = create_parser()
    args = parser.parse_args() if manually_provided_args is None else parser.parse_args(manually_provided_args)

    # No program specified, print help and exit
    if not args.module:
        parser.print_help()
        sys.exit()
    else:
        update_module(args=args)

    if testing:
        return args

    # call the relevant function
    args.func(args)

    return None
